#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""
This script checks that the pipfile of the repository meets the requirements.

In particular:
- Avoid the usage of "*"

It is assumed the script is run from the repository root.
"""

import itertools
import logging
import re
import sys
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional
from typing import OrderedDict as OrderedDictType
from typing import Tuple, cast

import click
import toml

from aea.configurations.data_types import Dependency
from aea.package_manager.base import load_configuration
from aea.package_manager.v1 import PackageManagerV1


ANY_SPECIFIER = "*"


class PathArgument(click.Path):
    """Path parameter for CLI."""

    def convert(
        self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Optional[Path]:
        """Convert path string to `pathlib.Path`"""
        path_string = super().convert(value, param, ctx)
        return None if path_string is None else Path(path_string)


class Pipfile:
    """Class to represent Pipfile config."""

    ignore = [
        "open-aea-ledger-cosmos",
        "open-aea-ledger-ethereum",
        "open-aea-ledger-fetchai",
        "open-aea-flashbots",
        "open-aea-flashbots",
        "tomte",
    ]

    def __init__(
        self,
        sources: List[str],
        packages: OrderedDictType[str, Dependency],
        dev_packages: OrderedDictType[str, Dependency],
        file: Path,
    ) -> None:
        """Initialize object."""
        self.sources = sources
        self.packages = packages
        self.dev_packages = dev_packages
        self.file = file

    def __iter__(self) -> Iterator[Dependency]:
        """Iterate dependencies as from aea.configurations.data_types.Dependency object."""
        for name, dependency in itertools.chain(
            self.packages.items(), self.dev_packages.items()
        ):
            if name.startswith("comment_") or name in self.ignore:
                continue
            yield dependency

    def update(self, dependency: Dependency) -> None:
        """Update dependency specifier"""
        if dependency.name in self.ignore:
            return
        if dependency.name in self.packages:
            if dependency.version == "":
                return
            self.packages[dependency.name] = dependency
        else:
            self.dev_packages[dependency.name] = dependency

    def check(self, dependency: Dependency) -> Tuple[Optional[str], int]:
        """Check dependency specifier"""
        if dependency.name in self.ignore:
            return None, 0

        if dependency.name in self.packages:
            expected = self.packages[dependency.name]
            if expected != dependency:
                return (
                    f"in Pipfile {expected.get_pip_install_args()[0]}; "
                    f"got {dependency.get_pip_install_args()[0]}"
                ), logging.WARNING
            return None, 0

        if dependency.name not in self.dev_packages:
            return f"{dependency.name} not found in Pipfile", logging.ERROR

        expected = self.dev_packages[dependency.name]
        if expected != dependency:
            return (
                f"in Pipfile {expected.get_pip_install_args()[0]}; "
                f"got {dependency.get_pip_install_args()[0]}"
            ), logging.WARNING

        return None, 0

    @classmethod
    def parse(
        cls, content: str
    ) -> Tuple[List[str], OrderedDictType[str, OrderedDictType[str, Dependency]]]:
        """Parse from string."""
        sources = []
        sections: OrderedDictType = OrderedDict()
        lines = content.split("\n")
        comments = 0
        while len(lines) > 0:
            line = lines.pop(0)
            if "[[source]]" in line:
                source = line + "\n"
                while True:
                    line = lines.pop(0)
                    if line == "":
                        break
                    source += line + "\n"
                sources.append(source)
            if "[dev-packages]" in line or "[packages]" in line:
                section = line
                sections[section] = OrderedDict()
                while len(lines) > 0:
                    line = lines.pop(0).strip()
                    if line == "":
                        break
                    if line.startswith("#"):
                        sections[section][f"comment_{comments}"] = line
                        comments += 1
                    else:
                        dep = Dependency.from_pipfile_string(line)
                        sections[section][dep.name] = dep
        return sources, sections

    def compile(self) -> str:
        """Compile to Pipfile string."""
        content = ""
        for source in self.sources:
            content += source + "\n"

        content += "[packages]\n"
        for package, dep in self.packages.items():
            if package.startswith("comment"):
                content += str(dep) + "\n"
            else:
                content += dep.to_pipfile_string() + "\n"

        content += "\n[dev-packages]\n"
        for package, dep in self.dev_packages.items():
            if package.startswith("comment"):
                content += str(dep) + "\n"
            else:
                content += dep.to_pipfile_string() + "\n"
        return content

    @classmethod
    def load(cls, file: Path) -> "Pipfile":
        """Load from file."""
        sources, sections = cls.parse(
            content=file.read_text(encoding="utf-8"),
        )
        return cls(
            sources=sources,
            packages=sections.get("[packages]", OrderedDict()),
            dev_packages=sections.get("[dev-packages]", OrderedDict()),
            file=file,
        )

    def dump(self) -> None:
        """Write to Pipfile."""
        self.file.write_text(self.compile(), encoding="utf-8")


class ToxFile:
    """Class to represent tox.ini file."""

    skip = [
        "open-aea-ledger-cosmos",
        "open-aea-ledger-ethereum",
        "open-aea-ledger-fetchai",
    ]

    def __init__(
        self,
        dependencies: Dict[str, Dict[str, Any]],
        file: Path,
    ) -> None:
        """Initialize object."""
        self.dependencies = dependencies
        self.file = file
        self.extra: Dict[str, Dependency] = {}

    def __iter__(self) -> Iterator[Dependency]:
        """Iter dependencies."""
        for obj in self.dependencies.values():
            yield obj["dep"]

    def update(self, dependency: Dependency) -> None:
        """Update dependency specifier"""
        if dependency.name in self.skip:
            return
        if dependency.name in self.dependencies:
            if dependency.version == "":
                return
            self.dependencies[dependency.name]["dep"] = dependency
            return
        self.extra[dependency.name] = dependency

    def check(self, dependency: Dependency) -> Tuple[Optional[str], int]:
        """Check dependency specifier"""
        if dependency.name in self.skip:
            return None, 0

        if dependency.name in self.dependencies:
            expected = self.dependencies[dependency.name]["dep"]
            if expected != dependency:
                return (
                    f"in tox.ini {expected.get_pip_install_args()[0]}; "
                    f"got {dependency.get_pip_install_args()[0]}"
                ), logging.WARNING
            return None, 0
        return f"{dependency.name} not found in tox.ini", logging.ERROR

    @classmethod
    def parse(cls, content: str) -> Dict[str, Dict[str, Any]]:
        """Parse file content."""
        deps = {}
        lines = content.split("\n")
        while len(lines) > 0:
            line = lines.pop(0)
            if line.startswith("deps"):
                while True:
                    line = lines.pop(0)
                    if not line.startswith("    "):
                        break
                    if (
                        line.startswith("    {")
                        or line.startswith("    ;")
                        or line.strip() == ""
                        or "tomte" in line
                    ):
                        continue
                    dep = Dependency.from_string(line.lstrip())
                    deps[dep.name] = {
                        "original": line,
                        "dep": dep,
                    }
        return deps

    @classmethod
    def load(cls, file: Path) -> "ToxFile":
        """Load tox.ini file."""
        content = file.read_text(encoding="utf-8")
        dependencies = cls.parse(content=content)
        return cls(
            dependencies=dependencies,
            file=file,
        )

    def _include_extra(self, content: str) -> str:
        """Include extra dependencies."""
        lines = content.split("\n")
        extra = []
        for dep in self.extra.values():
            extra.append(f"    {dep.get_pip_install_args()[0]}")

        if "[extra-deps]" in lines:
            start_idx = lines.index("[extra-deps]") + 2
            end_idx = lines.index("; end-extra")
            extra = list(sorted(set(extra + lines[start_idx:end_idx])))
            lines = lines[:start_idx] + extra + lines[end_idx:]
        else:
            idx = lines.index("[testenv]")
            lines = [
                *lines[:idx],
                "[extra-deps]",
                "deps = ",
                *list(sorted(extra)),
                "; end-extra\n",
                *lines[idx:],
            ]

        return "\n".join(lines)

    def write(self) -> None:
        """Dump config."""
        content = self.file.read_text(encoding="utf-8")
        for obj in self.dependencies.values():
            replace = "    " + cast(Dependency, obj["dep"]).get_pip_install_args()[0]
            content = re.sub(obj["original"], replace, content)

        if len(self.extra) > 0:
            content = self._include_extra(content=content)

        self.file.write_text(content, encoding="utf-8")


class PyProjectToml:
    """Class to represent pyproject.toml file."""

    ignore = [
        "python",
    ]

    def __init__(
        self,
        dependencies: OrderedDictType[str, Dependency],
        config: Dict[str, Dict],
        file: Path,
    ) -> None:
        """Initialize object."""
        self.dependencies = dependencies
        self.config = config
        self.file = file

    def __iter__(self) -> Iterator[Dependency]:
        """Iterate dependencies as from aea.configurations.data_types.Dependency object."""
        for dependency in self.dependencies.values():
            if dependency.name not in self.ignore:
                yield dependency

    def update(self, dependency: Dependency) -> None:
        """Update dependency specifier"""
        if dependency.name in self.ignore:
            return
        if dependency.name in self.dependencies and dependency.version == "":
            return
        self.dependencies[dependency.name] = dependency

    def check(self, dependency: Dependency) -> Tuple[Optional[str], int]:
        """Check dependency specifier"""
        if dependency.name in self.ignore:
            return None, 0

        if dependency.name not in self.dependencies:
            return f"{dependency.name} not found in pyproject.toml", logging.ERROR

        expected = self.dependencies[dependency.name]
        if expected != dependency:
            return (
                f"in pyproject.toml {expected.get_pip_install_args()[0]}; "
                f"got {dependency.get_pip_install_args()[0]}"
            ), logging.WARNING

        return None, 0

    @classmethod
    def load(cls, pyproject_path: Path) -> Optional["PyProjectToml"]:
        """Load pyproject.yaml dependencies"""
        config = toml.load(pyproject_path)
        dependencies = OrderedDict()
        try:
            config["tool"]["poetry"]["dependencies"]
        except KeyError:
            return None
        for name, version in config["tool"]["poetry"]["dependencies"].items():
            if isinstance(version, str):
                dependencies[name] = Dependency(
                    name=name,
                    version=version.replace("^", "==") if version != "*" else "",
                )
                continue
            data = cast(Dict, version)
            if "extras" in data:
                version = data["version"]
                if re.match(r"^\d", version):
                    version = f"=={version}"
                dependencies[name] = Dependency(
                    name=name,
                    version=version,
                    extras=data["extras"],
                )
                continue

        return cls(
            dependencies=dependencies,
            config=config,
            file=pyproject_path,
        )

    def dump(self) -> None:
        """Dump to file."""
        self.config["tool"]["poetry"]["dependencies"] = {
            package.name: package.version if package.version != "" else "*"
            for package in self.dependencies.values()
        }
        with self.file.open("w") as fp:
            toml.dump(self.config, fp)


def load_packages_dependencies(packages_dir: Path) -> List[Dependency]:
    """Returns a list of package dependencies."""
    package_manager = PackageManagerV1.from_dir(packages_dir=packages_dir)
    dependencies: Dict[str, Dependency] = {}
    for package in package_manager.iter_dependency_tree():
        if package.package_type.value == "service":
            continue
        _dependencies = load_configuration(  # type: ignore
            package_type=package.package_type,
            package_path=package_manager.package_path_from_package_id(
                package_id=package
            ),
        ).dependencies
        for key, value in _dependencies.items():
            if key not in dependencies:
                dependencies[key] = value
            else:
                if value.version == "":
                    continue
                if dependencies[key].version == "":
                    dependencies[key] = value
                if value == dependencies[key]:
                    continue
                print(
                    f"Non-matching dependency versions for {key}: {value} vs {dependencies[key]}"
                )

    return list(dependencies.values())


def _update(
    packages_dependencies: List[Dependency],
    tox: ToxFile,
    pipfile: Optional[Pipfile] = None,
    pyproject: Optional[PyProjectToml] = None,
) -> None:
    """Update dependencies."""

    if pipfile is not None:
        for dependency in packages_dependencies:
            pipfile.update(dependency=dependency)

        for dependency in pipfile:
            tox.update(dependency=dependency)

        for dependency in tox:
            pipfile.update(dependency=dependency)

        pipfile.dump()

    if pyproject is not None:
        for dependency in packages_dependencies:
            pyproject.update(dependency=dependency)

        for dependency in pyproject:
            tox.update(dependency=dependency)

        for dependency in tox:
            pyproject.update(dependency=dependency)

        pyproject.dump()

    tox.write()


def _check(
    packages_dependencies: List[Dependency],
    tox: ToxFile,
    pipfile: Optional[Pipfile] = None,
    pyproject: Optional[PyProjectToml] = None,
) -> None:
    """Update dependencies."""

    fail_check = 0

    if pipfile is not None:
        print("Comparing dependencies from Pipfile and packages")
        for dependency in packages_dependencies:
            error, level = pipfile.check(dependency=dependency)
            if error is not None:
                logging.log(level=level, msg=error)
                fail_check = level or fail_check

        print("Comparing dependencies from tox and Pipfile")
        for dependency in pipfile:
            error, level = tox.check(dependency=dependency)
            if error is not None:
                logging.log(level=level, msg=error)
                fail_check = level or fail_check

        print("Comparing dependencies from Pipfile and tox")
        for dependency in tox:
            error, level = pipfile.check(dependency=dependency)
            if error is not None:
                logging.log(level=level, msg=error)
                fail_check = level or fail_check

    if pyproject is not None:
        print("Comparing dependencies from pyproject.toml and packages")
        for dependency in packages_dependencies:
            error, level = pyproject.check(dependency=dependency)
            if error is not None:
                logging.log(level=level, msg=error)
                fail_check = level or fail_check

        print("Comparing dependencies from pyproject.toml and tox")
        for dependency in pyproject:
            error, level = tox.check(dependency=dependency)
            if error is not None:
                logging.log(level=level, msg=error)
                fail_check = level or fail_check

        print("Comparing dependencies from tox and pyproject.toml")
        for dependency in tox:
            error, level = pyproject.check(dependency=dependency)
            if error is not None:
                logging.log(level=level, msg=error)
                fail_check = level or fail_check

    print("Comparing dependencies from tox and packages")
    for dependency in packages_dependencies:
        error, level = tox.check(dependency=dependency)
        if error is not None:
            logging.log(level=level, msg=error)
            fail_check = level or fail_check

    if fail_check == logging.ERROR:
        print("Dependencies check failed")
        sys.exit(1)

    if fail_check == logging.WARNING:
        print("Please address warnings to avoid errors")
        sys.exit(0)

    print("No issues found")


@click.command(name="dm")
@click.option(
    "--check",
    is_flag=True,
    help="Perform dependency checks.",
)
@click.option(
    "--packages",
    "packages_dir",
    type=PathArgument(
        exists=True,
        file_okay=False,
        dir_okay=True,
    ),
    help="Path of the packages directory.",
)
@click.option(
    "--tox",
    "tox_path",
    type=PathArgument(
        exists=True,
        file_okay=True,
        dir_okay=False,
    ),
    help="Tox config path.",
)
@click.option(
    "--pipfile",
    "pipfile_path",
    type=PathArgument(
        exists=True,
        file_okay=True,
        dir_okay=False,
    ),
    help="Pipfile path.",
)
@click.option(
    "--pyproject",
    "pyproject_path",
    type=PathArgument(
        exists=True,
        file_okay=True,
        dir_okay=False,
    ),
    help="Pipfile path.",
)
def main(
    check: bool = False,
    packages_dir: Optional[Path] = None,
    tox_path: Optional[Path] = None,
    pipfile_path: Optional[Path] = None,
    pyproject_path: Optional[Path] = None,
) -> None:
    """Check dependencies across packages, tox.ini, pyproject.toml and setup.py"""

    logging.basicConfig(format="- %(levelname)s: %(message)s")

    tox_path = tox_path or Path.cwd() / "tox.ini"
    tox = ToxFile.load(tox_path)

    pipfile_path = pipfile_path or Path.cwd() / "Pipfile"
    pipfile = Pipfile.load(pipfile_path) if pipfile_path.exists() else None

    pyproject_path = pyproject_path or Path.cwd() / "pyproject.toml"
    pyproject = PyProjectToml.load(pyproject_path) if pyproject_path.exists() else None

    packages_dir = packages_dir or Path.cwd() / "packages"
    packages_dependencies = load_packages_dependencies(packages_dir=packages_dir)

    if check:
        return _check(
            tox=tox,
            pipfile=pipfile,
            pyproject=pyproject,
            packages_dependencies=packages_dependencies,
        )

    return _update(
        tox=tox,
        pipfile=pipfile,
        pyproject=pyproject,
        packages_dependencies=packages_dependencies,
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the `aea add connection` sub-command."""
import os
import shutil
import tempfile
import unittest.mock
from pathlib import Path
from unittest import mock

import pytest
import yaml
from jsonschema import ValidationError

import aea.configurations.base
from aea.cli import cli
from aea.cli.registry.settings import REMOTE_IPFS
from aea.configurations.base import DEFAULT_CONNECTION_CONFIG_FILE, PublicId
from aea.configurations.data_types import PackageId, PackageType
from aea.test_tools.test_cases import AEATestCaseEmpty, AEATestCaseEmptyFlaky

from packages.fetchai.connections.local.connection import (
    PUBLIC_ID as LOCAL_CONNECTION_PUBLIC_ID,
)
from packages.valory.connections.http_client.connection import (
    PUBLIC_ID as HTTP_CLIENT_PUBLIC_ID,
)

from tests.conftest import (
    AUTHOR,
    CLI_LOG_OPTION,
    CUR_PATH,
    CliRunner,
    MAX_FLAKY_RERUNS,
    TEST_IPFS_REGISTRY_CONFIG,
    double_escape_windows_path_separator,
    get_package_id_with_hash,
)
from tests.test_cli.test_add.test_generic import (
    BaseTestAddConnectionLocalWhenNoLocalRegistryExists,
    BaseTestAddRemoteMode,
    BaseTestAddSkillMixedModeFallsBack,
)


class TestAddConnectionFailsWhenConnectionAlreadyExists:
    """Test that the command 'aea add connection' fails when the connection already exists."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.connection_name = "http_client"
        cls.connection_author = "valory"
        cls.connection_version = "0.3.0"
        cls.connection_id = str(HTTP_CLIENT_PUBLIC_ID)
        # copy the 'packages' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "..", "packages"), Path(cls.t, "packages"))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR],
            standalone_mode=False,
        )

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)
        # add connection first time
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "connection", cls.connection_id],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        # add connection again
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "connection", cls.connection_id],
            standalone_mode=False,
        )

    @unittest.mock.patch("aea.cli.add.get_package_path", return_value="dest/path")
    @unittest.mock.patch(
        "aea.cli.add.get_default_remote_registry", return_value=REMOTE_IPFS
    )
    @unittest.mock.patch("aea.cli.add.fetch_ipfs")
    def test_add_connection_from_registry_positive(self, fetch_package_mock, *mocks):
        """Test add from registry positive result."""
        fetch_package_mock.return_value = Path(
            "vendor/{}/connections/{}".format(
                self.connection_author, self.connection_name
            )
        )
        public_id = "{}/{}:{}".format(
            AUTHOR, self.connection_name, self.connection_version
        )
        obj_type = "connection"
        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--remote", obj_type, public_id],
            standalone_mode=False,
        )
        assert result.exit_code == 0, result.stdout
        public_id_obj = PublicId.from_str(public_id)
        fetch_package_mock.assert_called_once_with(obj_type, public_id_obj, "dest/path")

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_error_message_connection_already_existing(self):
        """Test that the log error message is fixed.

        The expected message is: 'A connection with id '{connection_id}' already exists. Aborting...'
        """
        s = f"A connection with id '{self.connection_id}' already exists. Aborting..."
        assert self.result.exception.message == s

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestAddConnectionFailsWhenConnectionWithSameAuthorAndNameButDifferentVersion:
    """Test that 'aea add connection' fails when the connection with different version already exists."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.connection_name = "http_client"
        cls.connection_author = "valory"
        cls.connection_version = "0.3.0"
        cls.connection_id = str(HTTP_CLIENT_PUBLIC_ID)

        # copy the 'packages' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "..", "packages"), Path(cls.t, "packages"))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR],
            standalone_mode=False,
        )

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)
        # add connection first time
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "connection", cls.connection_id],
            standalone_mode=False,
        )
        assert result.exit_code == 0

        # add connection again, but with different version number
        # first, change version number to package
        different_version = "0.1.1"
        different_id = (
            cls.connection_author + "/" + cls.connection_name + ":" + different_version
        )
        config_path = Path(
            cls.t,
            "packages",
            cls.connection_author,
            "connections",
            cls.connection_name,
            DEFAULT_CONNECTION_CONFIG_FILE,
        )
        config = yaml.safe_load(config_path.open())
        config["version"] = different_version
        yaml.safe_dump(config, config_path.open(mode="w"))
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "connection", different_id],
            standalone_mode=False,
        )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_error_message_connection_already_existing(self):
        """Test that the log error message is fixed.

        The expected message is: 'A connection with id '{connection_id}' already exists. Aborting...'
        """
        s = f"A connection with id '{self.connection_id}' already exists. Aborting..."
        assert self.result.exception.message == s

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestAddConnectionFailsWhenConnectionNotInRegistry:
    """Test that the command 'aea add connection' fails when the connection is not in the registry."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.connection_id = "author/unknown_connection:0.1.0"
        cls.connection_name = "unknown_connection"

        # copy the 'packages' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "..", "packages"), Path(cls.t, "packages"))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR],
            standalone_mode=False,
        )

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "connection", cls.connection_id],
            standalone_mode=False,
        )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_error_message_connection_already_existing(self):
        """Test that the log error message is fixed.

        The expected message is: 'Cannot find connection: '{connection_name}''
        """
        s = "Cannot find connection: '{}'.".format(self.connection_id)
        assert self.result.exception.message == s

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestAddConnectionFailsWhenDifferentPublicId:
    """Test that the command 'aea add connection' fails when the connection has not the same public id."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.connection_id = "different_author/local:0.1.0"
        cls.connection_name = "unknown_connection"

        # copy the 'packages' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "..", "packages"), Path(cls.t, "packages"))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR],
            standalone_mode=False,
        )

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "connection", cls.connection_id],
            standalone_mode=False,
        )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_error_message_connection_wrong_public_id(self):
        """Test that the log error message is fixed."""
        s = "Cannot find connection: '{}'.".format(self.connection_id)
        assert self.result.exception.message == s

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestAddConnectionFailsWhenConfigFileIsNotCompliant:
    """Test that the command 'aea add connection' fails when the configuration file is not compliant with the schema."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.connection_id = str(HTTP_CLIENT_PUBLIC_ID)
        cls.connection_name = "http_client"

        # copy the 'packages' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "..", "packages"), Path(cls.t, "packages"))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR],
            standalone_mode=False,
        )

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0

        # change the serialization of the AgentConfig class so to make the parsing to fail.
        cls.patch = unittest.mock.patch.object(
            aea.configurations.base.ConnectionConfig,
            "from_json",
            side_effect=ValidationError("test error message"),
        )
        cls.patch.start()

        os.chdir(cls.agent_name)
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "connection", cls.connection_id],
            standalone_mode=False,
        )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_configuration_file_not_valid(self):
        """Test that the log error message is fixed.

        The expected message is: 'Connection configuration file not valid: '{connection_name}''
        """
        s = "Connection configuration file not valid: test error message"
        assert self.result.exception.message == s

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        cls.patch.stop()
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestAddConnectionFailsWhenDirectoryAlreadyExists:
    """Test that the command 'aea add connection' fails when the destination directory already exists."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.connection_id = str(HTTP_CLIENT_PUBLIC_ID)
        cls.connection_name = "http_client"

        # copy the 'packages' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "..", "packages"), Path(cls.t, "packages"))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR],
            standalone_mode=False,
        )

        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0

        os.chdir(cls.agent_name)
        Path(
            cls.t,
            cls.agent_name,
            "vendor",
            "valory",
            "connections",
            cls.connection_name,
        ).mkdir(parents=True, exist_ok=True)
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "add", "--local", "connection", cls.connection_id],
            standalone_mode=False,
        )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_file_exists_error(self):
        """Test that the log error message is fixed.

        The expected message is: 'Cannot find connection: '{connection_name}''
        """
        missing_path = os.path.join(
            "vendor", "valory", "connections", self.connection_name
        )
        missing_path = double_escape_windows_path_separator(missing_path)
        assert missing_path in self.result.exception.message

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


@pytest.mark.integration
class TestAddConnectionFromRemoteRegistryHTTP(AEATestCaseEmptyFlaky):
    """Test case for add connection from Registry command."""

    IS_LOCAL = False
    IS_EMPTY = True

    @pytest.mark.integration
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    @mock.patch(
        "aea.cli.registry.utils.get_or_create_cli_config",
        return_value=TEST_IPFS_REGISTRY_CONFIG,
    )
    @mock.patch("aea.cli.add.get_default_remote_registry", return_value=REMOTE_IPFS)
    @mock.patch("aea.cli.add.is_fingerprint_correct", return_value=True)
    def test_add_connection_from_remote_registry_positive(self, *_):
        """Test add connection from Registry positive result."""
        self.add_item(
            "connection",
            str(LOCAL_CONNECTION_PUBLIC_ID.to_latest()),
            local=self.IS_LOCAL,
        )

        items_path = os.path.join(self.agent_name, "vendor", "fetchai", "connections")
        items_folders = os.listdir(items_path)
        item_name = "local"
        assert item_name in items_folders


class TestAddConnectionWithLatestVersion(AEATestCaseEmpty):
    """Test case for add connection with latest version."""

    def test_add_connection_latest_version(self):
        """Test add connection with latest version."""
        self.add_item(
            "connection", str(LOCAL_CONNECTION_PUBLIC_ID.to_latest()), local=True
        )

        items_path = os.path.join(self.agent_name, "vendor", "fetchai", "connections")
        items_folders = os.listdir(items_path)
        item_name = "local"
        assert item_name in items_folders


class TestAddConnectionMixedModeFallsBack(BaseTestAddSkillMixedModeFallsBack):
    """Test add skill in mixed mode that fails with local falls back to remote registry."""

    COMPONENT_ID = get_package_id_with_hash(
        PackageId(
            package_type=PackageType.CONNECTION,
            public_id=PublicId(
                "valory",
                "http_client",
                "0.23.0",
            ),
        )
    ).public_id
    COMPONENT_TYPE = PackageType.CONNECTION


class TestAddConnectionLocalWhenNoLocalRegistryExists(
    BaseTestAddConnectionLocalWhenNoLocalRegistryExists
):
    """Test that the command 'aea add connection' fails in local mode when the local registry does not exists."""

    COMPONENT_ID = HTTP_CLIENT_PUBLIC_ID
    COMPONENT_TYPE = PackageType.CONNECTION


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
class TestAddConnectionMode(BaseTestAddRemoteMode):
    """Test case for add connection, --remote mode."""

    COMPONENT_ID = TestAddConnectionMixedModeFallsBack.COMPONENT_ID
    COMPONENT_TYPE = PackageType.CONNECTION



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for fingerprinting packages."""

import random
import shutil
import tempfile
from pathlib import Path

import pytest
import yaml

from aea.configurations.base import PACKAGE_TYPE_TO_CONFIG_CLASS
from aea.helpers.cid import CID
from aea.helpers.dependency_tree import COMPONENTS
from aea.helpers.fingerprint import (
    check_fingerprint,
    compute_fingerprint,
    update_fingerprint,
)

from tests.conftest import PACKAGES_DIR


CONFIG_CLASSES = {k.value: v for k, v in PACKAGE_TYPE_TO_CONFIG_CLASS.items()}
CONFIG_FILES = {k: Path(PACKAGES_DIR).rglob(f_name) for k, f_name in COMPONENTS}


def test_compute_fingerprint():
    """Test compute_fingerprint"""

    ignore_pattern = "__init__.py"
    package_path = Path(PACKAGES_DIR)
    fingerprints = compute_fingerprint(package_path, fingerprint_ignore_patterns=None)
    assert all(map(lambda multihash: CID.is_cid(multihash), fingerprints.values()))

    n_fingerprints_without_ignore = len(fingerprints)
    n_init_dot_py = sum(p.endswith(ignore_pattern) for p in fingerprints)
    fingerprints = compute_fingerprint(package_path, (ignore_pattern,))
    assert len(fingerprints) + n_init_dot_py == n_fingerprints_without_ignore
    assert all(map(lambda multihash: CID.is_cid(multihash), fingerprints.values()))


@pytest.mark.parametrize("package_type, files", CONFIG_FILES.items())
def test_update_fingerprint(package_type, files):
    """Test update fingerprint"""

    def load_config(file_path):
        """Load config.yaml"""
        # we ignore overwrites below `---` in this test
        # tests dealing with loading are under `test_configurations`
        file_contents = file_path.read_text().split("---")
        base_doc = file_contents[0]
        json = yaml.safe_load(base_doc)
        return config_cls._create_or_update_from_json(json)

    def point_mutation(directory: Path) -> None:
        """Mutate random character in a random python file"""
        py_file = random.choice(list(directory.glob("*.py")))  # nosec
        content = Path(py_file).read_text()
        i = random.choice(range(len(content)))  # nosec
        new = content[:i] + chr(ord(content[i]) + 1) + content[i + 1 :]
        py_file.write_text(new)

    dir_none_error_msg = "configuration.directory cannot be None."
    config_cls = CONFIG_CLASSES.get(package_type)
    with tempfile.TemporaryDirectory() as tmp_dir:
        for file in files:
            # copy files and operate on those
            nested_dir = Path(tmp_dir, file.parent.parts[-1])
            shutil.copytree(file.parent, nested_dir)
            copied_file = nested_dir / file.name

            config = load_config(copied_file)
            with pytest.raises(ValueError, match=dir_none_error_msg):
                update_fingerprint(config)

            # no code accompanies the aea-config.yaml file
            # check agent returns True before directory check
            if package_type == "agent":
                assert check_fingerprint(config)
                shutil.rmtree(nested_dir)
                continue

            # check all other types of packages
            with pytest.raises(ValueError, match=dir_none_error_msg):
                check_fingerprint(config)
            config.directory = nested_dir

            assert check_fingerprint(config)
            original_fp = config.fingerprint
            point_mutation(nested_dir)
            assert not check_fingerprint(config)
            update_fingerprint(config)
            config = load_config(copied_file)
            config.directory = nested_dir
            assert check_fingerprint(config)
            assert not original_fp == config.fingerprint

            shutil.rmtree(nested_dir)

            # test fingerprint empty
            shutil.copy(file, tmp_dir)
            config = load_config(Path(tmp_dir) / file.name)
            assert config.fingerprint
            config.directory = Path(tmp_dir)
            update_fingerprint(config)
            config = load_config(Path(tmp_dir) / file.name)
            assert not config.fingerprint



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Custom implementation of `eth_account.Account` for hardware wallets."""

from contextlib import contextmanager
from itertools import islice
from typing import Any, Generator, List, NamedTuple, Optional

import rlp
from aea_ledger_ethereum_hwi.bip32 import h, m
from aea_ledger_ethereum_hwi.exceptions import HWIError
from construct import (
    Byte,
    Bytes,
    BytesInteger,
    GreedyBytes,
    Int8ub,
    Int32ub,
    PascalString,
    Prefixed,
    PrefixedArray,
    Struct,
)
from cytoolz import dissoc
from eth_account._utils.legacy_transactions import (
    encode_transaction,
    serializable_unsigned_transaction_from_dict,
)
from eth_account._utils.typed_transactions import TypedTransaction
from eth_account.datastructures import HexBytes, SignedMessage, SignedTransaction
from eth_account.messages import SignableMessage
from eth_keys.main import PublicKey
from eth_rlp import HashableRLP
from eth_typing.evm import ChecksumAddress
from eth_utils.curried import keccak
from ledgerwallet.client import CommException, LedgerClient
from ledgerwallet.transport import enumerate_devices

from aea.common import JSONLike


address = rlp.sedes.Binary.fixed_length(20, allow_empty=True)
access_list_sede_type = rlp.sedes.CountableList(
    rlp.sedes.List(
        [
            address,
            rlp.sedes.CountableList(rlp.sedes.BigEndianInt(32)),
        ]
    ),
)


def chunk(seq: bytes, size: int) -> List[bytes]:
    """Converts a byte sequence to a list of chunks"""
    it = iter(seq)
    chunks = list(iter(lambda: tuple(islice(it, size)), ()))
    return [bytes(each) for each in chunks]


class HWIAccountData(NamedTuple):
    """Hardware wallet account data"""

    public_key: str
    address: str
    chain_code: Optional[bytes]


class HWISignedTransaction(NamedTuple):
    """Hardware wallet signed transaction"""

    v: int
    r: int
    s: int


class UnsignedDynamicTransaction(HashableRLP):
    """Unsigned dynamic transaction."""

    fields = [
        ("chain_id", rlp.sedes.big_endian_int),
        ("nonce", rlp.sedes.big_endian_int),
        ("max_priority_fee_per_gas", rlp.sedes.big_endian_int),
        ("max_fee_per_gas", rlp.sedes.big_endian_int),
        ("gas_limit", rlp.sedes.big_endian_int),
        ("destination", address),
        ("amount", rlp.sedes.big_endian_int),
        ("data", rlp.sedes.binary),
        ("access_list", access_list_sede_type),
    ]


class UnsignedType1Transaction(HashableRLP):
    """Unsigned typ1 transaction"""

    fields = [
        ("chainId", rlp.sedes.big_endian_int),
        ("nonce", rlp.sedes.big_endian_int),
        ("gasPrice", rlp.sedes.big_endian_int),
        ("gas", rlp.sedes.big_endian_int),
        ("to", address),
        ("value", rlp.sedes.big_endian_int),
        ("data", rlp.sedes.binary),
        ("accessList", access_list_sede_type),
    ]


class SignTransactionAPDU:
    """Sign transaction APDU codes"""

    INS = 0x04
    P1_0 = 0x00
    P1 = 0x80
    P2 = 0x00


class GetAccountAPDU:
    """Get account APDU codes"""

    INS = 0x02
    P1 = 0x01
    P2 = 0x01


SignedTransactionStruct = Struct(
    v=BytesInteger(1),
    r=BytesInteger(32),
    s=BytesInteger(32),
)


AccountStruct = Struct(
    public_key=Prefixed(Int8ub, GreedyBytes),
    address=PascalString(Int8ub, "ascii"),
    chain_code=Bytes(32),
)


class HWIErrorCodes:
    """HWI com errors."""

    DEVICE_LOCKED = 21781
    EHTEREUM_APP_NOT_OPEN = 25870


@contextmanager
def reraise_from_hwi_com_error() -> Generator:
    """Reraise ledger communication exception as `HWIError`"""

    try:
        yield
    except CommException as e:
        if e.sw == HWIErrorCodes.DEVICE_LOCKED:
            raise HWIError(message="Device is locked", sw=e.sw)
        elif e.sw == HWIErrorCodes.EHTEREUM_APP_NOT_OPEN:
            raise HWIError(message="Please open ethereum app in your device", sw=e.sw)
        else:
            raise HWIError(message=e.message, sw=e.sw, data=e.data)


class HWIAccount:
    """Hardware wallet interface as ethereum account similar to `eth_account.Account` to represent `Crypto.entity`"""

    default_account: Optional[HWIAccountData]

    def __init__(
        self,
        default_device_index: int = 0,
        default_account_index: int = 0,
        default_key_index: int = 0,
    ) -> None:
        """Initialize object."""

        self._default_device_index = default_device_index
        self._default_key_index = default_key_index
        self._default_account_index = default_account_index

        self.default_account = None

    @property
    def devices(self) -> List[Any]:
        """Returns the list of available devices."""
        devices = enumerate_devices()
        if len(devices) == 0:
            raise HWIError(message="Cannot find any ledger device", sw=0)
        return devices

    def get_client(
        self,
        device_index: Optional[int] = None,
    ) -> LedgerClient:
        """Get ledger client."""
        device_index = device_index or self._default_device_index
        return LedgerClient(device=self.devices[device_index])

    def get_account(
        self,
        key_index: Optional[int] = None,
        account_index: Optional[int] = None,
        device_index: Optional[int] = None,
    ) -> HWIAccountData:
        """Get hardware wallet account."""

        key_index = key_index or self._default_key_index
        account_index = account_index or self._default_account_index
        path = m / h(44) / h(60) / h(account_index) / 0 / key_index

        path_construct = PrefixedArray(Byte, Int32ub)
        path_apdu = path_construct.build(path.to_list())
        client = self.get_client(device_index=device_index)

        with reraise_from_hwi_com_error():
            response = client.apdu_exchange(
                ins=GetAccountAPDU.INS,
                data=path_apdu,
                p1=GetAccountAPDU.P1,
                p2=GetAccountAPDU.P2,
            )

        parsed_response = AccountStruct.parse(response)
        pbk = PublicKey(parsed_response.public_key[1:])

        return HWIAccountData(
            public_key=str(pbk),
            address=pbk.to_checksum_address(),
            chain_code=parsed_response.chain_code,
        )

    @property
    def address(self) -> ChecksumAddress:
        """Address"""
        if self.default_account is None:
            self.default_account = self.get_account()
        return self.default_account.address

    @property
    def public_key(self) -> ChecksumAddress:
        """Public key"""
        if self.default_account is None:
            self.default_account = self.get_account()
        return self.default_account.public_key

    def sign_message(
        self, signable_message: SignableMessage, **kwargs: Any
    ) -> SignedMessage:
        """Sign a EIP191 message"""

        raise NotImplementedError()

    @staticmethod
    def encode_transaction(
        transaction: TypedTransaction,
        is_eip1559_tx: bool = False,
        key_index: Optional[int] = None,
        account_index: Optional[int] = None,
    ) -> bytes:
        """Build and encode transaction"""

        account_index = account_index or 0
        key_index = key_index or 0

        # BIP44 Path m/44'/cointype'/account'/change/address
        path = m / h(44) / h(60) / h(account_index) / 0 / key_index
        path_construct = PrefixedArray(Byte, Int32ub)
        path_apdu = path_construct.build(path.to_list())

        if is_eip1559_tx:
            as_dict = transaction.as_dict()
            tx = UnsignedDynamicTransaction.from_dict(
                field_dict=dict(
                    chain_id=as_dict["chainId"],
                    nonce=as_dict["nonce"],
                    max_priority_fee_per_gas=as_dict["maxPriorityFeePerGas"],
                    max_fee_per_gas=as_dict["maxFeePerGas"],
                    gas_limit=as_dict.get("gas", 21000),
                    destination=as_dict["to"],
                    amount=as_dict["value"],
                    data=as_dict["data"],
                    access_list=as_dict.get("accessList", []),
                )
            )
            return (
                path_apdu
                + transaction.transaction_type.to_bytes(1, "big")
                + rlp.encode(tx)
            )

        as_dict = transaction.as_dict()
        tx = UnsignedType1Transaction.from_dict(
            field_dict=dict(
                chainId=as_dict["chainId"],
                nonce=as_dict["nonce"],
                gasPrice=as_dict["gasPrice"],
                gas=as_dict["gas"],
                to=as_dict["to"],
                value=as_dict["value"],
                data=as_dict["data"],
                accessList=as_dict.get("accessList", []),
            )
        )
        return (
            path_apdu + transaction.transaction_type.to_bytes(1, "big") + rlp.encode(tx)
        )

    def _sign_transaction_on_hwi(
        self,
        unsigned_transaction: TypedTransaction,
        key_index: Optional[int] = None,
        account_index: Optional[int] = None,
        device_index: Optional[int] = None,
    ) -> HWISignedTransaction:
        """Hardware wallet signed transaction"""

        client = self.get_client(
            device_index=device_index,
        )

        is_eip1559_tx = (
            hasattr(unsigned_transaction, "transaction_type")
            and unsigned_transaction.transaction_type == 2
        )

        request_data = self.encode_transaction(
            transaction=unsigned_transaction,
            is_eip1559_tx=is_eip1559_tx,
            key_index=key_index,
            account_index=account_index,
        )

        with reraise_from_hwi_com_error():
            raw_response = bytes()
            for idx, each in enumerate(chunk(request_data, 255)):
                raw_response = client.apdu_exchange(
                    ins=SignTransactionAPDU.INS,
                    data=each,
                    p1=(
                        SignTransactionAPDU.P1_0 if idx == 0 else SignTransactionAPDU.P1
                    ),
                    p2=SignTransactionAPDU.P2,
                )
        parsed_response = SignedTransactionStruct.parse(raw_response)

        return HWISignedTransaction(
            v=parsed_response.v,
            r=parsed_response.r,
            s=parsed_response.s,
        )

    def sign_transaction(
        self,
        transaction_dict: JSONLike,
        **kwargs: Any,
    ) -> SignedTransaction:
        """Sign transaction."""

        key_index: Optional[int] = kwargs.get("key_index")
        device_index: Optional[int] = kwargs.get("device_index")
        account_index: Optional[int] = kwargs.get("account_index")

        account = self.get_account(
            key_index=key_index, account_index=account_index, device_index=device_index
        )
        self.default_account = account

        chain_id = transaction_dict.get(
            "chainId",
            transaction_dict.get(
                "chain_id",
                1,
            ),
        )

        if "from" in transaction_dict:
            if transaction_dict["from"].lower() == account.address.lower():
                sanitized_transaction = dissoc(transaction_dict, "from")
            else:
                raise TypeError(
                    "from field must match key's %s, but it was %s"
                    % (
                        account.address,
                        transaction_dict["from"],
                    )
                )
        else:
            sanitized_transaction = transaction_dict

        to_address = sanitized_transaction["to"]
        if isinstance(to_address, str):
            sanitized_transaction["to"] = bytes.fromhex(to_address[2:])

        if isinstance(to_address, bytes) and to_address.startswith(b"0x"):
            sanitized_transaction["to"] = to_address[2:]

        if (
            "maxPriorityFeePerGas" in sanitized_transaction
            or "max_priority_fee_per_gas" in sanitized_transaction
        ):
            sanitized_transaction["type"] = 2
        else:
            sanitized_transaction["type"] = 1

        transaction_dict["chainId"] = chain_id

        # encode transaction
        unsigned_transaction = serializable_unsigned_transaction_from_dict(
            sanitized_transaction
        )
        transaction_hash = unsigned_transaction.hash()

        response = self._sign_transaction_on_hwi(
            unsigned_transaction=unsigned_transaction,
            device_index=device_index,
            key_index=key_index,
            account_index=account_index,
        )

        # serialize transaction with rlp
        encoded_transaction = encode_transaction(
            unsigned_transaction,
            vrs=(
                response.v,
                response.r,
                response.s,
            ),
        )
        transaction_hash = keccak(encoded_transaction)

        return SignedTransaction(
            rawTransaction=HexBytes(encoded_transaction),
            hash=HexBytes(transaction_hash),
            r=response.r,
            s=response.s,
            v=response.v,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests for the `aea add connection` sub-command."""
# import filecmp
# import os
# import shutil
# import tempfile
# from contextlib import contextmanager
# from pathlib import Path
# from typing import List, Set, cast
# from unittest import mock
# from unittest.mock import MagicMock, patch

# import pytest
# from click.exceptions import ClickException
# from click.testing import Result
# from packaging.version import Version

# import aea
# from aea import get_current_aea_version
# from aea.cli import cli
# from aea.cli.registry.utils import get_latest_version_available_in_registry
# from aea.cli.upgrade import ItemRemoveHelper
# from aea.cli.utils.config import load_item_config
# from aea.configurations.base import (
#     AgentConfig,
#     ComponentId,
#     ComponentType,
#     DEFAULT_AEA_CONFIG_FILE,
#     PackageId,
#     PackageType,
#     PublicId,
# )
# from aea.configurations.constants import DEFAULT_VERSION
# from aea.configurations.loader import ConfigLoader, load_component_configuration
# from aea.helpers.base import cd, compute_specifier_from_version
# from aea.test_tools.test_cases import AEATestCaseEmpty, BaseAEATestCase

# from packages.fetchai.connections import local
# from packages.fetchai.connections.oef.connection import PUBLIC_ID as OEF_PUBLIC_ID
# from packages.fetchai.connections.soef.connection import PUBLIC_ID as SOEF_PUBLIC_ID
# from packages.fetchai.connections.stub.connection import StubConnection
# from packages.fetchai.contracts.erc1155.contract import PUBLIC_ID as ERC1155_PUBLIC_ID
# from packages.fetchai.protocols.default import DefaultMessage
# from packages.valory.protocols.http.message import HttpMessage
# from packages.fetchai.protocols.oef_search.message import OefSearchMessage
# from packages.fetchai.skills.echo import PUBLIC_ID as ECHO_SKILL_PUBLIC_ID
# from packages.fetchai.skills.error import PUBLIC_ID as ERROR_SKILL_PUBLIC_ID

# from tests.common.mocks import RegexComparator
# from tests.common.utils import are_dirs_equal, dircmp_recursive
# from tests.conftest import AUTHOR, CLI_LOG_OPTION, CUR_PATH, CliRunner


# class BaseTestCase:
#     """Base test case class with setup and teardown and some utils."""

#     ITEM_TYPE = "connection"
#     ITEM_PUBLIC_ID = SOEF_PUBLIC_ID
#     LOCAL: List[str] = ["--local"]
#     DEPENDENCY_TYPE = "protocol"
#     DEPENDENCY_PUBLIC_ID = OefSearchMessage.protocol_id

#     @staticmethod
#     def loader() -> ConfigLoader:
#         """Return Agent config loader."""
#         return ConfigLoader.from_configuration_type(PackageType.AGENT)

#     def load_config(self) -> AgentConfig:
#         """Load AgentConfig from current directory."""
#         agent_loader = self.loader()
#         path = Path(DEFAULT_AEA_CONFIG_FILE)
#         with path.open(mode="r", encoding="utf-8") as fp:
#             agent_config = agent_loader.load(fp)
#         return agent_config

#     def load_mock_context(self) -> MagicMock:
#         """Load mock context."""
#         context_mock = MagicMock(agent_config=self.load_config())
#         return context_mock

#     def dump_config(self, agent_config: AgentConfig) -> None:
#         """Dump AgentConfig to current directory."""
#         agent_loader = self.loader()
#         path = Path(DEFAULT_AEA_CONFIG_FILE)

#         with path.open(mode="w", encoding="utf-8") as fp:
#             agent_loader.dump(agent_config, fp)

#     @classmethod
#     def setup(cls):
#         """Set the test up."""
#         cls.runner = CliRunner()
#         cls.agent_name = "myagent"
#         cls.cwd = os.getcwd()
#         cls.t = tempfile.mkdtemp()
#         # copy the 'packages' directory in the parent of the agent folder.
#         shutil.copytree(Path(CUR_PATH, "..", "packages"), Path(cls.t, "packages"))

#         os.chdir(cls.t)
#         result = cls.runner.invoke(
#             cli,
#             [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR, "--default-registry", "http"],
#             standalone_mode=False,
#         )
#         assert result.exit_code == 0
#         result = cls.runner.invoke(
#             cli,
#             [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
#             standalone_mode=False,
#         )
#         assert result.exit_code == 0
#         os.chdir(cls.agent_name)
#         # add connection first time

#     @contextmanager
#     def with_oef_installed(self):
#         """Add and remove oef connection."""
#         result = self.runner.invoke(
#             cli,
#             [
#                 "-v",
#                 "DEBUG",
#                 "add",
#                 "--local",
#                 "connection",
#                 str(oef.connection.PUBLIC_ID),
#             ],
#             standalone_mode=False,
#         )
#         assert result.exit_code == 0
#         try:
#             yield
#         finally:
#             result = self.runner.invoke(
#                 cli,
#                 ["-v", "DEBUG", "remove", "connection", str(oef.connection.PUBLIC_ID)],
#                 standalone_mode=False,
#             )
#             assert result.exit_code == 0

#     @contextmanager
#     def with_config_update(self):
#         """Context manager to update item version to 0.0.1."""
#         original_config = self.load_config()

#         config_data = original_config.json
#         if str(self.ITEM_PUBLIC_ID) in config_data[f"{self.ITEM_TYPE}s"]:
#             config_data[f"{self.ITEM_TYPE}s"].remove(str(self.ITEM_PUBLIC_ID))
#         config_data[f"{self.ITEM_TYPE}s"].append(
#             f"{self.ITEM_PUBLIC_ID.author}/{self.ITEM_PUBLIC_ID.name}:0.0.1"
#         )
#         self.dump_config(AgentConfig.from_json(config_data))
#         try:
#             yield
#         finally:
#             self.dump_config(original_config)

#     @classmethod
#     def teardown(cls):
#         """Tear the test down."""
#         os.chdir(cls.cwd)
#         try:
#             shutil.rmtree(cls.t)
#         except (OSError, IOError):
#             pass


# class TestRemoveAndDependencies(BaseTestCase):
#     """Test dependency remove helper and upgrade with dependency removed."""

#     ITEM_TYPE = "connection"
#     ITEM_PUBLIC_ID = SOEF_PUBLIC_ID
#     LOCAL: List[str] = ["--local"]
#     DEPENDENCY_TYPE = "protocol"
#     DEPENDENCY_PUBLIC_ID = OefSearchMessage.protocol_id

#     @classmethod
#     def setup(cls):
#         """Set the test up."""
#         super(TestRemoveAndDependencies, cls).setup()
#         cls.DEPENDENCY_PACKAGE_ID = PackageId(
#             cls.DEPENDENCY_TYPE, cls.DEPENDENCY_PUBLIC_ID
#         )
#         result = cls.runner.invoke(
#             cli,
#             ["-v", "DEBUG", "add", "--local", cls.ITEM_TYPE, str(cls.ITEM_PUBLIC_ID)],
#             standalone_mode=False,
#         )
#         assert result.exit_code == 0

#     def test_upgrade_and_dependency_removed(self):
#         """
#         Test dependency removed after upgrade.

#         Done with mocking _add_item_deps to avoid dependencies installation.

#         Also checks dependency configuration removed with component
#         """
#         assert self.DEPENDENCY_PUBLIC_ID in self.load_config().protocols

#         # add empty component config to aea-config.py
#         agent_config = self.load_config()
#         component_id = ComponentId(self.DEPENDENCY_TYPE, self.DEPENDENCY_PUBLIC_ID)
#         agent_config.component_configurations[component_id] = {}  # just empty
#         agent_config.component_configurations[
#             ComponentId(self.ITEM_TYPE, self.ITEM_PUBLIC_ID)
#         ] = {}  # just empty
#         self.dump_config(agent_config)

#         agent_config = self.load_config()
#         assert component_id in agent_config.component_configurations

#         with patch(
#             "aea.cli.upgrade.ItemUpgrader.check_upgrade_is_required",
#             return_value=self.ITEM_PUBLIC_ID.version,
#         ), patch("aea.cli.add._add_item_deps"):
#             result = self.runner.invoke(
#                 cli,
#                 [
#                     "-v",
#                     "DEBUG",
#                     "upgrade",
#                     *self.LOCAL,
#                     self.ITEM_TYPE,
#                     f"{self.ITEM_PUBLIC_ID.author}/{self.ITEM_PUBLIC_ID.name}:latest",
#                 ],
#                 catch_exceptions=True,
#             )
#             try:
#                 assert result.exit_code == 0

#                 assert self.DEPENDENCY_PUBLIC_ID not in self.load_config().protocols
#                 agent_config = self.load_config()

#                 # check configuration was removed too
#                 assert component_id not in agent_config.component_configurations
#                 assert (
#                     ComponentId(self.ITEM_TYPE, self.ITEM_PUBLIC_ID)
#                     in agent_config.component_configurations
#                 )
#             finally:
#                 # restore component removed
#                 result = self.runner.invoke(
#                     cli,
#                     [
#                         "-v",
#                         "DEBUG",
#                         "add",
#                         *self.LOCAL,
#                         self.DEPENDENCY_TYPE,
#                         f"{self.DEPENDENCY_PUBLIC_ID.author}/{self.DEPENDENCY_PUBLIC_ID.name}:latest",
#                     ],
#                     catch_exceptions=True,
#                 )
#                 assert self.DEPENDENCY_PUBLIC_ID in self.load_config().protocols

#     def test_upgrade_and_dependency_not_removed_caused_required_by_another_item(self):
#         """Test dependency is not removed after upgrade cause required by another item."""
#         assert self.DEPENDENCY_PUBLIC_ID in self.load_config().protocols
#         # do not add dependencies for the package

#         with self.with_oef_installed(), self.with_config_update(), patch(
#             "aea.cli.add._add_item_deps"
#         ):
#             result = self.runner.invoke(
#                 cli,
#                 [
#                     "-v",
#                     "DEBUG",
#                     "upgrade",
#                     *self.LOCAL,
#                     self.ITEM_TYPE,
#                     f"{self.ITEM_PUBLIC_ID.author}/{self.ITEM_PUBLIC_ID.name}:latest",
#                 ],
#                 catch_exceptions=True,
#             )
#             assert result.exit_code == 0
#             assert self.DEPENDENCY_PUBLIC_ID in self.load_config().protocols


# class TestUpgradeSharedDependencies(AEATestCaseEmpty):
#     """
#     Test removal of shared dependency.

#     The shared dependency in this test case is 'fetchai/oef_search:0.9.0'.
#     """

#     IS_EMPTY = True
#     OLD_SOEF_ID = PublicId.from_str("fetchai/soef:0.16.0")
#     OLD_OEF_SEARCH_ID = PublicId.from_str("fetchai/oef_search:0.13.0")
#     OLD_OEF_ID = PublicId.from_str("fetchai/oef:0.16.0")

#     @classmethod
#     def setup_class(cls):
#         """
#         Set the test up.

#         Skip consistency checks to avoid aea version compatibility checks.
#         """
#         super().setup_class()
#         result = cls.run_cli_command(
#             "-s", "add", "connection", str(cls.OLD_SOEF_ID), cwd=cls._get_cwd()
#         )
#         assert result.exit_code == 0
#         result = cls.run_cli_command(
#             "-s", "add", "connection", str(cls.OLD_OEF_ID), cwd=cls._get_cwd()
#         )
#         assert result.exit_code == 0

#     def test_upgrade_shared_dependencies(self):
#         """Test upgrade shared dependencies."""
#         result = self.run_cli_command("-s", "upgrade", cwd=self._get_cwd())
#         assert result.exit_code == 0

#         agent_config: AgentConfig = cast(
#             AgentConfig,
#             load_item_config(PackageType.AGENT.value, Path(self.current_agent_context)),
#         )
#         assert OefSearchMessage.protocol_id in agent_config.protocols
#         assert SOEF_PUBLIC_ID in agent_config.connections
#         assert OEF_PUBLIC_ID in agent_config.connections


# class TestUpgradeProject(BaseAEATestCase, BaseTestCase):
#     """Test that the command 'aea upgrade' works."""

#     capture_log = True

#     @classmethod
#     def setup(cls):
#         """Set up test case."""
#         super(TestUpgradeProject, cls).setup()
#         cls.change_directory(Path(".."))
#         cls.agent_name = "generic_buyer"
#         cls.latest_agent_name = "generic_buyer_latest"
#         cls.run_cli_command(
#             "--skip-consistency-check",
#             "fetch",
#             "fetchai/generic_buyer:0.29.0",
#             "--alias",
#             cls.agent_name,
#         )
#         cls.run_cli_command(
#             "--skip-consistency-check",
#             "fetch",
#             "fetchai/generic_buyer:latest",
#             "--alias",
#             cls.latest_agent_name,
#         )
#         cls.agents.add(cls.agent_name)
#         cls.set_agent_context(cls.agent_name)

#     def test_upgrade(self):
#         """Test upgrade project old version to latest one and compare with latest project fetched."""
#         with cd(self.latest_agent_name):
#             latest_agent_items = set(
#                 ItemRemoveHelper(self.load_mock_context())
#                 .get_agent_dependencies_with_reverse_dependencies()
#                 .keys()
#             )

#         with cd(self.agent_name):
#             self.runner.invoke(  # pylint: disable=no-member
#                 cli,
#                 ["--skip-consistency-check", "upgrade", "--local"],
#                 standalone_mode=False,
#                 catch_exceptions=False,
#             )
#             agent_items = set(
#                 ItemRemoveHelper(self.load_mock_context())
#                 .get_agent_dependencies_with_reverse_dependencies()
#                 .keys()
#             )
#             assert latest_agent_items == agent_items

#         # upgrade again to check it workd with upgraded version
#         with cd(self.agent_name):
#             self.runner.invoke(  # pylint: disable=no-member
#                 cli,
#                 ["--skip-consistency-check", "upgrade", "--local"],
#                 standalone_mode=False,
#                 catch_exceptions=False,
#             )
#             agent_items = set(
#                 ItemRemoveHelper(self.load_mock_context())
#                 .get_agent_dependencies_with_reverse_dependencies()
#                 .keys()
#             )
#             assert latest_agent_items == agent_items

#         # compare both configuration files, except the agent name and the author
#         upgraded_agent_dir = Path(self.agent_name)
#         latest_agent_dir = Path(self.latest_agent_name)
#         lines_upgraded_agent_config = (
#             (upgraded_agent_dir / DEFAULT_AEA_CONFIG_FILE).read_text().splitlines()
#         )
#         lines_latest_agent_config = (
#             (latest_agent_dir / DEFAULT_AEA_CONFIG_FILE).read_text().splitlines()
#         )
#         # the slice is because we don't compare the agent name and the author name
#         assert lines_upgraded_agent_config[2:] == lines_latest_agent_config[2:]

#         # compare vendor folders.
#         assert are_dirs_equal(
#             upgraded_agent_dir / "vendor", latest_agent_dir / "vendor"
#         )


# class TestNonVendorProject(BaseAEATestCase, BaseTestCase):
#     """Test that the command 'aea upgrade' works."""

#     capture_log = True

#     @classmethod
#     def setup(cls):
#         """Set up test case."""
#         super(TestNonVendorProject, cls).setup()
#         cls.change_directory(Path(".."))
#         cls.agent_name = "generic_buyer"
#         cls.run_cli_command(
#             "fetch", "fetchai/generic_buyer:0.29.0", "--alias", cls.agent_name
#         )
#         cls.agents.add(cls.agent_name)
#         cls.set_agent_context(cls.agent_name)

#     @patch("aea.cli.upgrade.ItemUpgrader.is_non_vendor", True)
#     @patch(
#         "aea.cli.upgrade.ItemUpgrader.check_upgrade_is_required", return_value="0.99.0"
#     )
#     @patch("aea.cli.upgrade.ItemUpgrader.remove_item")
#     @patch("aea.cli.upgrade.ItemUpgrader.add_item")
#     def test_non_vendor_nothing_to_upgrade(
#         self, *mocks
#     ):  # pylint: disable=unused-argument
#         """Test upgrade project dependencies not removed cause non vendor."""
#         with cd(self.agent_name):
#             base_agent_items = set(
#                 ItemRemoveHelper(self.load_mock_context())
#                 .get_agent_dependencies_with_reverse_dependencies()
#                 .keys()
#             )

#             self.runner.invoke(  # pylint: disable=no-member
#                 cli,
#                 ["--skip-consistency-check", "upgrade"],
#                 standalone_mode=False,
#                 catch_exceptions=False,
#             )
#             agent_items = set(
#                 ItemRemoveHelper(self.load_mock_context())
#                 .get_agent_dependencies_with_reverse_dependencies()
#                 .keys()
#             )
#             assert base_agent_items == agent_items


# class TestUpgradeConnectionLocally(BaseTestCase):
#     """Test that the command 'aea upgrade connection' works."""

#     ITEM_TYPE = "connection"
#     ITEM_PUBLIC_ID = SOEF_PUBLIC_ID
#     LOCAL: List[str] = ["--local"]

#     @classmethod
#     def setup(cls):
#         """Set the test up."""
#         super(TestUpgradeConnectionLocally, cls).setup()

#         result = cls.runner.invoke(
#             cli,
#             ["-v", "DEBUG", "add", "--local", cls.ITEM_TYPE, str(cls.ITEM_PUBLIC_ID)],
#             standalone_mode=False,
#         )
#         assert result.exit_code == 0

#     def test_upgrade_to_same_version(self):
#         """Test do not  upgrade to version already installed."""
#         with pytest.raises(
#             ClickException,
#             match=r"The .* with id '.*' already has version .*. Nothing to upgrade.",
#         ):
#             self.runner.invoke(
#                 cli,
#                 ["upgrade", *self.LOCAL, self.ITEM_TYPE, str(self.ITEM_PUBLIC_ID)],
#                 standalone_mode=False,
#                 catch_exceptions=False,
#             )

#     @patch("aea.cli.upgrade.ItemUpgrader.is_non_vendor", True)
#     def test_upgrade_non_vendor(self):
#         """Test do not upgrade non vendor package."""
#         with pytest.raises(
#             ClickException,
#             match=r"The .* with id '.*' already has version .*. Nothing to upgrade.",
#         ):
#             self.runner.invoke(
#                 cli,
#                 [
#                     "upgrade",
#                     *self.LOCAL,
#                     self.ITEM_TYPE,
#                     f"{self.ITEM_PUBLIC_ID.author}/{self.ITEM_PUBLIC_ID.name}:100.0.0",
#                 ],
#                 standalone_mode=False,
#                 catch_exceptions=False,
#             )

#     def test_upgrade_to_latest_but_same_version(self):
#         """Test no update to latest if already latest component."""
#         with pytest.raises(
#             ClickException,
#             match=r"The .* with id '.*' already has version .*. Nothing to upgrade.",
#         ):
#             self.runner.invoke(
#                 cli,
#                 [
#                     "upgrade",
#                     *self.LOCAL,
#                     self.ITEM_TYPE,
#                     f"{self.ITEM_PUBLIC_ID.author}/{self.ITEM_PUBLIC_ID.name}:latest",
#                 ],
#                 standalone_mode=False,
#                 catch_exceptions=False,
#             )

#     def test_upgrade_to_non_registered(self):
#         """Test can not upgrade not registered component."""
#         with pytest.raises(
#             ClickException,
#             match=r".* with id .* is not registered. Please use the `add` command. Aborting...",
#         ):
#             self.runner.invoke(
#                 cli,
#                 [
#                     "-v",
#                     "DEBUG",
#                     "upgrade",
#                     *self.LOCAL,
#                     self.ITEM_TYPE,
#                     "nonexits/dummy:0.0.0",
#                 ],
#                 standalone_mode=False,
#                 catch_exceptions=False,
#             )

#     def test_upgrade_required_mock(self):
#         """Test upgrade with mocking upgrade required."""
#         with patch(
#             "aea.cli.upgrade.ItemUpgrader.check_upgrade_is_required",
#             return_value="100.0.0",
#         ):
#             result = self.runner.invoke(
#                 cli,
#                 [
#                     "-v",
#                     "DEBUG",
#                     "upgrade",
#                     *self.LOCAL,
#                     self.ITEM_TYPE,
#                     f"{self.ITEM_PUBLIC_ID.author}/{self.ITEM_PUBLIC_ID.name}:latest",
#                 ],
#                 catch_exceptions=False,
#             )
#             assert result.exit_code == 0

#     def test_do_upgrade(self):
#         """Test real full upgrade."""
#         with self.with_config_update():
#             result = self.runner.invoke(
#                 cli,
#                 [
#                     "upgrade",
#                     *self.LOCAL,
#                     self.ITEM_TYPE,
#                     f"{self.ITEM_PUBLIC_ID.author}/{self.ITEM_PUBLIC_ID.name}:latest",
#                 ],
#                 standalone_mode=False,
#             )
#             assert result.exit_code == 0

#     def test_package_can_not_be_found_in_registry(self):
#         """Test no package in registry."""
#         with self.with_config_update():
#             with patch(
#                 "aea.cli.registry.utils.get_package_meta",
#                 side_effects=Exception("expected!"),
#             ), patch(
#                 "aea.cli.registry.utils.find_item_locally",
#                 side_effects=Exception("expected!"),
#             ), pytest.raises(
#                 ClickException,
#                 match=r"Package .* details can not be fetched from the registry!",
#             ):
#                 self.runner.invoke(
#                     cli,
#                     [
#                         "upgrade",
#                         *self.LOCAL,
#                         self.ITEM_TYPE,
#                         f"{self.ITEM_PUBLIC_ID.author}/{self.ITEM_PUBLIC_ID.name}:latest",
#                     ],
#                     standalone_mode=False,
#                     catch_exceptions=False,
#                 )

#     def test_package_can_not_upgraded_cause_required(self):
#         """Test no package in registry."""
#         with self.with_config_update():
#             with patch(
#                 "aea.cli.upgrade.ItemRemoveHelper.check_remove",
#                 return_value=(
#                     set([PackageId("connection", PublicId("test", "test", "0.0.1"))]),
#                     set(),
#                     dict(),
#                 ),
#             ), pytest.raises(
#                 ClickException,
#                 match=r"Can not upgrade .* because it is required by '.*'",
#             ):
#                 self.runner.invoke(
#                     cli,
#                     [
#                         "upgrade",
#                         *self.LOCAL,
#                         self.ITEM_TYPE,
#                         f"{self.ITEM_PUBLIC_ID.author}/{self.ITEM_PUBLIC_ID.name}:latest",
#                     ],
#                     standalone_mode=False,
#                     catch_exceptions=False,
#                 )

#     @classmethod
#     def teardown(cls):
#         """Tear the test down."""
#         super(TestUpgradeConnectionLocally, cls).teardown()

#         os.chdir(cls.cwd)
#         try:
#             shutil.rmtree(cls.t)
#         except (OSError, IOError):
#             pass


# class TestUpgradeConnectionRemoteRegistry(TestUpgradeConnectionLocally):
#     """Test that the command 'aea upgrade connection' works."""

#     LOCAL: List[str] = []

#     def test_upgrade_to_latest_but_same_version(self):
#         """Skip."""
#         pass


# class TestUpgradeProtocolLocally(TestUpgradeConnectionLocally):
#     """Test that the command 'aea upgrade protocol --local' works."""

#     ITEM_TYPE = "protocol"
#     ITEM_PUBLIC_ID = HttpMessage.protocol_id


# class TestUpgradeProtocolRemoteRegistry(TestUpgradeProtocolLocally):
#     """Test that the command 'aea upgrade protocol' works."""

#     LOCAL: List[str] = []

#     def test_upgrade_to_latest_but_same_version(self):
#         """Skip."""
#         pass


# class TestUpgradeSkillLocally(TestUpgradeConnectionLocally):
#     """Test that the command 'aea upgrade skill --local' works."""

#     ITEM_TYPE = "skill"
#     ITEM_PUBLIC_ID = ECHO_SKILL_PUBLIC_ID


# class TestUpgradeSkillRemoteRegistry(TestUpgradeSkillLocally):
#     """Test that the command 'aea upgrade skill' works."""

#     LOCAL: List[str] = []

#     def test_upgrade_to_latest_but_same_version(self):
#         """Skip."""
#         pass

#     def test_upgrade_required_mock(self):
#         """Skip."""
#         pass

#     def test_do_upgrade(self):
#         """Skip."""
#         pass


# class TestUpgradeContractLocally(TestUpgradeConnectionLocally):
#     """Test that the command 'aea upgrade contract' works."""

#     ITEM_TYPE = "contract"
#     ITEM_PUBLIC_ID = ERC1155_PUBLIC_ID


# class TestUpgradeContractRemoteRegistry(TestUpgradeContractLocally):
#     """Test that the command 'aea upgrade contract --local' works."""

#     LOCAL: List[str] = []

#     def test_upgrade_to_latest_but_same_version(self):
#         """Skip."""
#         pass


# @pytest.mark.integration
# class TestUpgradeNonVendorDependencies(AEATestCaseEmpty):
#     """
#     Test that the command 'aea upgrade' correctly updates non-vendor package data.

#     In particular, check that 'aea upgrade' updates:
#     - the public ids of the package dependencies and the 'aea_version' field.
#     - the 'aea_version' field in case it is not compatible with the current version.

#     The test works as follows:
#     - scaffold a package, one for each possible package type;
#     - add the protocol "fetchai/default:0.12.0" as dependency to each of them.
#     - add the skill "fetchai/error:0.12.0"; this will also add the default protocol.
#       add it also as dependency of non-vendor skill.
#     - run 'aea upgrade'
#     - check that the reference to "fetchai/default" in each scaffolded package
#       has the new version.
#     """

#     capture_log = True
#     IS_EMPTY = True
#     old_default_protocol_id = PublicId(
#         DefaultMessage.protocol_id.author, DefaultMessage.protocol_id.name, "0.12.0"
#     )
#     old_error_skill_id = PublicId(
#         ERROR_SKILL_PUBLIC_ID.author, ERROR_SKILL_PUBLIC_ID.name, "0.12.0"
#     )
#     old_aea_version_range = compute_specifier_from_version(Version("0.1.0"))

#     @classmethod
#     def scaffold_item(
#         cls, item_type: str, name: str, skip_consistency_check: bool = False
#     ) -> Result:
#         """Override default behaviour by adding a custom dependency to the scaffolded item."""
#         result = super(TestUpgradeNonVendorDependencies, cls).scaffold_item(
#             item_type, name, skip_consistency_check
#         )
#         # add custom dependency (a protocol) to each package
#         # that supports dependencies (only connections and skills)
#         if item_type in {ComponentType.CONNECTION.value, ComponentType.SKILL.value}:
#             cls.nested_set_config(
#                 f"{ComponentType(item_type).to_plural()}.{name}.protocols",
#                 [str(cls.old_default_protocol_id)],
#             )

#         # add the vendor skill as dependency of the non-vendor skill
#         if item_type == ComponentType.SKILL.value:
#             cls.nested_set_config(
#                 f"{ComponentType(item_type).to_plural()}.{name}.skills",
#                 [str(cls.old_error_skill_id)],
#             )

#         # update 'aea_version' to an old version range.
#         cls.nested_set_config(
#             f"{ComponentType(item_type).to_plural()}.{name}.aea_version",
#             str(cls.old_aea_version_range),
#         )
#         return result

#     @classmethod
#     def setup_class(cls):
#         """Set up test case."""
#         super(TestUpgradeNonVendorDependencies, cls).setup_class()
#         cls.scaffold_item("protocol", "my_protocol", skip_consistency_check=True)
#         cls.scaffold_item("connection", "my_connection", skip_consistency_check=True)
#         cls.scaffold_item("contract", "my_contract", skip_consistency_check=True)
#         cls.scaffold_item("skill", "my_skill", skip_consistency_check=True)
#         cls.run_cli_command(
#             "--skip-consistency-check",
#             "add",
#             "skill",
#             str(cls.old_error_skill_id),
#             cwd=cls._get_cwd(),
#         )
#         cls.run_cli_command(
#             "--skip-consistency-check", "upgrade", "--local", cwd=cls._get_cwd()
#         )

#     def test_agent_config_updated(self):
#         """Test the agent configuration is updated."""
#         loader = ConfigLoader.from_configuration_type(PackageType.AGENT)
#         with Path(self._get_cwd(), DEFAULT_AEA_CONFIG_FILE).open() as fp:
#             agent_config = loader.load(fp)
#         assert DefaultMessage.protocol_id in agent_config.protocols
#         assert ERROR_SKILL_PUBLIC_ID in agent_config.skills

#     def test_non_vendor_update_references_to_upgraded_packages(
#         self,
#     ):  # pylint: disable=unused-argument
#         """Test that dependencies in non-vendor packages are updated correctly after upgrade."""
#         self.assert_dependency_updated(
#             ComponentType.CONNECTION,
#             "my_connection",
#             "protocols",
#             {DefaultMessage.protocol_id},
#         )
#         self.assert_dependency_updated(
#             ComponentType.SKILL, "my_skill", "protocols", {DefaultMessage.protocol_id}
#         )
#         self.assert_dependency_updated(
#             ComponentType.SKILL, "my_skill", "skills", {ERROR_SKILL_PUBLIC_ID}
#         )

#     def assert_dependency_updated(
#         self,
#         item_type: ComponentType,
#         package_name: str,
#         package_type: str,
#         expected: Set[PublicId],
#     ):
#         """Assert dependency is updated."""
#         package_path = Path(self._get_cwd(), item_type.to_plural(), package_name)
#         component_config = load_component_configuration(item_type, package_path)
#         assert hasattr(component_config, package_type), "Test is not well-written."
#         assert getattr(component_config, package_type) == expected  # type: ignore

#         expected_version_range = compute_specifier_from_version(
#             get_current_aea_version()
#         )
#         assert component_config.aea_version == expected_version_range


# class TestUpdateReferences(AEATestCaseEmpty):
#     """
#     Test that references are updated correctly after 'aea upgrade'.

#     In particular, 'default_routing', 'default_connection' and custom component configurations in AEA configuration.

#     How the test works:
#     - add fetchai/error:0.12.0, that requires fetchai/default:0.12.0
#     - add fetchai/stub:0.16.0
#     - add 'fetchai/default:0.12.0: fetchai/stub:0.16.0' to default routing
#     - add custom configuration to stub connection.
#     - run 'aea upgrade'. This will upgrade `stub` connection and `error` skill, and in turn `default` protocol.
#     """

#     IS_EMPTY = True

#     OLD_DEFAULT_PROTOCOL_PUBLIC_ID = PublicId.from_str("fetchai/default:0.12.0")
#     OLD_ERROR_SKILL_PUBLIC_ID = PublicId.from_str("fetchai/error:0.12.0")
#     OLD_STUB_CONNECTION_PUBLIC_ID = PublicId.from_str("fetchai/stub:0.16.0")

#     @classmethod
#     def setup_class(cls):
#         """Set up the test class."""
#         super().setup_class()
#         cls.run_cli_command(
#             "--skip-consistency-check",
#             "add",
#             "skill",
#             str(cls.OLD_ERROR_SKILL_PUBLIC_ID),
#             cwd=cls._get_cwd(),
#         )
#         cls.run_cli_command(
#             "--skip-consistency-check",
#             "add",
#             "connection",
#             str(cls.OLD_STUB_CONNECTION_PUBLIC_ID),
#             cwd=cls._get_cwd(),
#         )

#         cls.nested_set_config(
#             "agent.default_routing",
#             {cls.OLD_DEFAULT_PROTOCOL_PUBLIC_ID: cls.OLD_STUB_CONNECTION_PUBLIC_ID},
#         )
#         cls.nested_set_config(
#             "agent.default_connection", cls.OLD_STUB_CONNECTION_PUBLIC_ID,
#         )
#         cls.run_cli_command(
#             "--skip-consistency-check",
#             "config",
#             "set",
#             "vendor.fetchai.skills.error.is_abstract",
#             "--type",
#             "bool",
#             "true",
#             cwd=cls._get_cwd(),
#         )

#         cls.run_cli_command(
#             "--skip-consistency-check", "upgrade", "--local", cwd=cls._get_cwd()
#         )

#     def test_default_routing_updated_correctly(self):
#         """Test default routing has been updated correctly."""
#         result = self.run_cli_command(
#             "--skip-consistency-check",
#             "config",
#             "get",
#             "agent.default_routing",
#             cwd=self._get_cwd(),
#         )
#         assert (
#             result.stdout
#             == f'{{"{DefaultMessage.protocol_id}": "{StubConnection.connection_id}"}}\n'
#         )

#     def test_default_connection_updated_correctly(self):
#         """Test default routing has been updated correctly."""
#         result = self.run_cli_command(
#             "--skip-consistency-check",
#             "config",
#             "get",
#             "agent.default_connection",
#             cwd=self._get_cwd(),
#         )
#         assert result.stdout == "fetchai/stub:0.21.0\n"

#     def test_custom_configuration_updated_correctly(self):
#         """Test default routing has been updated correctly."""
#         result = self.run_cli_command(
#             "--skip-consistency-check",
#             "config",
#             "get",
#             "vendor.fetchai.skills.error.is_abstract",
#             cwd=self._get_cwd(),
#         )
#         assert result.stdout == "True\n"


# @mock.patch("click.echo")
# class TestNothingToUpgrade(AEATestCaseEmpty):
#     """Test the upgrade command when there's nothing t upgrade."""

#     def test_nothing_to_upgrade(self, mock_click_echo):
#         """Test nothing to upgrade."""
#         agent_config = self.load_agent_config(self.agent_name)
#         result = self.run_cli_command("upgrade", cwd=self._get_cwd())
#         assert result.exit_code == 0
#         mock_click_echo.assert_any_call("Starting project upgrade...")
#         mock_click_echo.assert_any_call(
#             f"Checking if there is a newer remote version of agent package '{agent_config.public_id}'..."
#         )
#         mock_click_echo.assert_any_call(
#             "Package not found, continuing with normal upgrade."
#         )
#         mock_click_echo.assert_any_call("Everything is already up to date!")


# @pytest.mark.integration
# @mock.patch("click.echo")
# class TestWrongAEAVersion(AEATestCaseEmpty):
#     """
#     Test consistency check ignores AEA version fields.

#     Use an old version of a package to simulate an upgrade.
#     """

#     AEA_VERSION_SPECIFIER: str = "==0.1.0"
#     IS_EMPTY = True

#     @classmethod
#     def setup_class(cls):
#         """Set up the test."""
#         super().setup_class()

#         # this is an old version of the package, just to trigger an upgrade.
#         cls.add_item("protocol", "fetchai/default:0.12.0", local=False)

#         # change aea version of the AEA project
#         agent_config = cls.load_agent_config(cls.current_agent_context)
#         cls._update_aea_version(agent_config)
#         cls.nested_set_config("agent.aea_version", cls.AEA_VERSION_SPECIFIER)
#         cls.nested_set_config("agent.author", "wrong_author")

#     @classmethod
#     def _update_aea_version(cls, agent_config: AgentConfig):
#         """Update aea version to all items and fingerprint them."""
#         for item in agent_config.package_dependencies:
#             type_ = item.component_type.to_plural()
#             dotted_path = f"vendor.{item.author}.{type_}.{item.name}.aea_version"
#             path = os.path.join("vendor", item.author, type_, item.name)
#             cls.nested_set_config(dotted_path, cls.AEA_VERSION_SPECIFIER)
#             cls.run_cli_command("fingerprint", "by-path", path, cwd=cls._get_cwd())

#     def test_nothing_to_upgrade(self, mock_click_echo):
#         """Test nothing to upgrade, and additionally, that 'aea_version' is correct."""
#         result = self.run_cli_command("upgrade", cwd=self._get_cwd())
#         assert result.exit_code == 0
#         mock_click_echo.assert_any_call("Starting project upgrade...")
#         mock_click_echo.assert_any_call(
#             f"Updating AEA version specifier from ==0.1.0 to {compute_specifier_from_version(get_current_aea_version())}."
#         )

#         # test 'aea_version' of agent configuration is upgraded
#         expected_aea_version_specifier = compute_specifier_from_version(
#             get_current_aea_version()
#         )
#         agent_config = self.load_agent_config(self.current_agent_context)
#         assert agent_config.aea_version == expected_aea_version_specifier
#         assert agent_config.author == self.author
#         assert agent_config.version == DEFAULT_VERSION


# @mock.patch("click.echo")
# @mock.patch("click.confirm")
# @mock.patch("aea.cli.upgrade.get_latest_version_available_in_registry")
# class BaseTestUpgradeWithEject(AEATestCaseEmpty):
#     """
#     Base test class to test 'aea upgrade' with request for ejection.

#     We use an old version of 'generic seller' skill to simulate a request from the CLI tool.
#     The utility 'get_latest_version_available_in_registry' is mocked so to
#     hide the new version of that package, hence triggering an ejection.;
#     """

#     IS_EMPTY = True

#     GENERIC_SELLER = ComponentId(
#         ComponentType.SKILL, PublicId.from_str("fetchai/generic_seller:0.27.0")
#     )
#     unmocked = get_latest_version_available_in_registry

#     EXPECTED_CLICK_ECHO_CALLS: List[str] = []
#     EXPECTED_CLICK_CONFIRM_CALLS: List[str] = []
#     CONFIRM_OUTPUT = [False, False]

#     @classmethod
#     def setup_class(cls):
#         """Set up the test."""
#         super().setup_class()
#         cls.add_item("skill", str(cls.GENERIC_SELLER.public_id), local=False)

#     @classmethod
#     def mock_get_latest_version_available_in_registry(cls, *args, **kwargs):
#         """Mock 'get_latest_version_available_in_registry' when called with generic_seller public key."""
#         if (
#             args[1] == str(cls.GENERIC_SELLER.package_type)
#             and args[2] == cls.GENERIC_SELLER.public_id.to_latest()
#         ):
#             # return current version
#             return cls.GENERIC_SELLER
#         return cls.unmocked(*args, **kwargs)

#     def _get_mock(self):
#         """Get the mock of 'get_latest_version_available_in_registry'."""
#         return mock.patch(
#             "aea.cli.upgrade.get_latest_version_available_in_registry",
#             side_effect=self.mock_get_latest_version_available_in_registry,
#         )

#     def test_run(self, mock_get_latest_version, mock_click_confirm, mock_click_echo):
#         """Run the test."""
#         mock_get_latest_version.side_effect = (
#             self.mock_get_latest_version_available_in_registry
#         )
#         mock_click_confirm.side_effect = self.CONFIRM_OUTPUT
#         result = self.run_cli_command("upgrade", cwd=self._get_cwd())
#         assert result.exit_code == 0
#         self.mock_click_echo = mock_click_echo
#         self.mock_click_confirm = mock_click_confirm
#         self._assert_calls(self.EXPECTED_CLICK_ECHO_CALLS, mock_click_echo)
#         self._assert_calls(self.EXPECTED_CLICK_CONFIRM_CALLS, mock_click_confirm)

#     def _assert_calls(self, args: List, stdout_mock: MagicMock):
#         """Assert lines are present in stdout."""
#         for expected_line in args:
#             stdout_mock.assert_any_call(expected_line)


# @pytest.mark.integration
# class TestUpgradeWithEjectAbort(BaseTestUpgradeWithEject):
#     """Test 'aea upgrade' command with request for ejection, refused."""

#     GENERIC_SELLER = ComponentId(
#         ComponentType.SKILL, PublicId.from_str("fetchai/generic_seller:0.24.0")
#     )

#     EXPECTED_CLICK_ECHO_CALLS = ["Abort."]
#     EXPECTED_CLICK_CONFIRM_CALLS = [
#         RegexComparator(
#             r"Skill fetchai/generic_seller:0.24.0 prevents the upgrade of the following vendor packages:.*as there isn't a compatible version available on the AEA registry\. Would you like to eject it\?"
#         )
#     ]


# @pytest.mark.integration
# class TestUpgradeWithEjectAccept(BaseTestUpgradeWithEject):
#     """Test 'aea upgrade' command with request for ejection, accepted by the user."""

#     CONFIRM_OUTPUT = [True, True]

#     GENERIC_SELLER = ComponentId(
#         ComponentType.SKILL, PublicId.from_str("fetchai/generic_seller:0.24.0")
#     )

#     EXPECTED_CLICK_ECHO_CALLS = [
#         "Ejecting (skill, fetchai/generic_seller:0.24.0)...",
#         "Ejecting item skill fetchai/generic_seller:0.24.0",
#         "Fingerprinting skill components of 'default_author/generic_seller:0.1.0' ...",
#         "Successfully ejected skill fetchai/generic_seller:0.24.0 to ./skills/generic_seller as default_author/generic_seller:0.1.0.",
#     ]
#     EXPECTED_CLICK_CONFIRM_CALLS = [
#         RegexComparator(
#             "Skill fetchai/generic_seller:0.24.0 prevents the upgrade of the following vendor packages:"
#         ),
#         RegexComparator(
#             "as there isn't a compatible version available on the AEA registry. Would you like to eject it?"
#         ),
#     ]

#     def test_run(self, *mocks):
#         """Run the test."""
#         super().test_run(*mocks)
#         ejected_package_path = Path(
#             self.t, self.current_agent_context, "skills", "generic_seller"
#         )
#         assert ejected_package_path.exists()
#         assert ejected_package_path.is_dir()


# @pytest.mark.integration
# class BaseTestUpgradeProject(AEATestCaseEmpty):
#     """Base test class for testing project upgrader."""

#     OLD_AGENT_PUBLIC_ID = PublicId.from_str("fetchai/weather_station:0.26.0")
#     EXPECTED_NEW_AGENT_PUBLIC_ID = OLD_AGENT_PUBLIC_ID.to_latest()
#     EXPECTED = "expected_agent"

#     @classmethod
#     def setup_class(cls):
#         """Set up the test."""
#         super().setup_class()
#         cls.run_cli_command(
#             "--skip-consistency-check",
#             "fetch",
#             "--remote",
#             str(cls.EXPECTED_NEW_AGENT_PUBLIC_ID),
#             "--alias",
#             cls.EXPECTED,
#         )

#     def setup(self):
#         """Set up the class."""
#         self.run_cli_command("fetch", str(self.OLD_AGENT_PUBLIC_ID))
#         self.set_agent_context(self.OLD_AGENT_PUBLIC_ID.name)

#     def teardown(self):
#         """Tear down class."""
#         shutil.rmtree(self.current_agent_context)


# @mock.patch("click.confirm")
# class TestUpgradeProjectWithNewerVersion(BaseTestUpgradeProject):
#     """Test upgrade project with newer version available."""

#     @pytest.mark.parametrize("confirm", [True, False])
#     def test_upgrade(self, mock_confirm, confirm):
#         """Test upgrade."""
#         mock_confirm.return_value = confirm
#         result = self.run_cli_command("upgrade", "--remote", cwd=self._get_cwd())
#         assert result.exit_code == 0

#         mock_confirm.assert_any_call(
#             RegexComparator(
#                 r"Found a newer version of this project:.*Would you like to replace this project with it\?.*Warning: the content in the current directory.*will be removed"
#             )
#         )

#         # compare with latest fetched agent.
#         ignore = [DEFAULT_AEA_CONFIG_FILE] + filecmp.DEFAULT_IGNORES
#         dircmp = filecmp.dircmp(
#             self.current_agent_context, self.EXPECTED, ignore=ignore
#         )
#         _left_only, _right_only, diff = dircmp_recursive(dircmp)
#         assert _right_only == diff == _left_only == set()


# @mock.patch("aea.cli.upgrade.get_latest_version_available_in_registry")
# @mock.patch("click.echo")
# class TestUpgradeProjectWithoutNewerVersion(BaseTestUpgradeProject):
#     """Test upgrade project without newer version available (but available on registry)."""

#     def test_run(self, mock_click_echo, mock_get_latest_version):
#         """Run the test."""
#         fake_old_public_id = self.OLD_AGENT_PUBLIC_ID
#         mock_get_latest_version.return_value = fake_old_public_id
#         result = self.run_cli_command("upgrade", "--remote", cwd=self._get_cwd())
#         assert result.exit_code == 0

#         version_str = str(self.OLD_AGENT_PUBLIC_ID.version)
#         mock_click_echo.assert_any_call(
#             f"Latest version found is '{version_str}' which is smaller or equal than current version '{version_str}'. Continuing..."
#         )

#         # compare with latest fetched agent.
#         ignore = [DEFAULT_AEA_CONFIG_FILE] + filecmp.DEFAULT_IGNORES
#         dircmp = filecmp.dircmp(
#             self.current_agent_context, self.EXPECTED, ignore=ignore
#         )
#         _left_only, _right_only, diff = dircmp_recursive(dircmp)
#         assert diff == set()
#         # temp: assert diff == _left_only == _right_only == set()


# @mock.patch.object(aea, "__version__", "0.11.0")
# class TestUpgradeAEACompatibility(BaseTestUpgradeProject):
#     """
#     Test 'aea upgrade' takes into account the current aea version.

#     The test works as follows:
#     """

#     OLD_AGENT_PUBLIC_ID = PublicId.from_str("fetchai/weather_station:0.31.0")
#     EXPECTED_NEW_AGENT_PUBLIC_ID = PublicId.from_str("fetchai/weather_station:latest")

#     def test_upgrade(self):
#         """Test upgrade."""
#         result = self.run_cli_command("upgrade", "--remote", "-y", cwd=self._get_cwd())
#         assert result.exit_code == 0

#         # compare with latest fetched agent.
#         ignore = [DEFAULT_AEA_CONFIG_FILE] + filecmp.DEFAULT_IGNORES
#         dircmp = filecmp.dircmp(
#             self.current_agent_context, self.EXPECTED, ignore=ignore
#         )
#         _left_only, _right_only, diff = dircmp_recursive(dircmp)
#         assert diff == set()  # temp

#         # compare agent configuration files (except the name)
#         expected_content = (
#             Path(self.EXPECTED, DEFAULT_AEA_CONFIG_FILE).read_text().splitlines()[1:]
#         )
#         actual_content = (
#             Path(self.current_agent_context, DEFAULT_AEA_CONFIG_FILE)
#             .read_text()
#             .splitlines()[1:]
#         )
#         assert expected_content != actual_content  # temp



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test messages module for contract_api protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import List

from aea.test_tools.test_protocol import BaseProtocolMessagesTestCase

from packages.valory.protocols.contract_api.custom_types import (
    Kwargs,
    RawMessage,
    RawTransaction,
    State,
)
from packages.valory.protocols.contract_api.message import ContractApiMessage


class TestMessageContractApi(BaseProtocolMessagesTestCase):
    """Test for the 'contract_api' protocol message."""

    MESSAGE_CLASS = ContractApiMessage

    def build_messages(self) -> List[ContractApiMessage]:  # type: ignore[override]
        """Build the messages to be used for testing."""
        return [
            ContractApiMessage(
                performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
                ledger_id="some str",
                contract_id="some str",
                callable="some str",
                kwargs=Kwargs({"key_1": 1, "key_2": 2}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
                ledger_id="some str",
                contract_id="some str",
                contract_address="some str",
                callable="some str",
                kwargs=Kwargs({"key_1": 1, "key_2": 2}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.GET_RAW_MESSAGE,
                ledger_id="some str",
                contract_id="some str",
                contract_address="some str",
                callable="some str",
                kwargs=Kwargs({"key_1": 1, "key_2": 2}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.GET_STATE,
                ledger_id="some str",
                contract_id="some str",
                contract_address="some str",
                callable="some str",
                kwargs=Kwargs({"key_1": 1, "key_2": 2}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.STATE,
                state=State("some_ledger_id", {"key": "some_body"}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                raw_transaction=RawTransaction("some_ledger_id", {"body": "some_body"}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.RAW_MESSAGE,
                raw_message=RawMessage("some_ledger_id", b"some_body"),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.ERROR,
                code=12,
                message="some str",
                data=b"some_bytes",
            ),
        ]

    def build_inconsistent(self) -> List[ContractApiMessage]:  # type: ignore[override]
        """Build inconsistent messages to be used for testing."""
        return [
            ContractApiMessage(
                performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
                # skip content: ledger_id
                contract_id="some str",
                callable="some str",
                kwargs=Kwargs({"key_1": 1, "key_2": 2}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
                # skip content: ledger_id
                contract_id="some str",
                contract_address="some str",
                callable="some str",
                kwargs=Kwargs({"key_1": 1, "key_2": 2}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.GET_RAW_MESSAGE,
                # skip content: ledger_id
                contract_id="some str",
                contract_address="some str",
                callable="some str",
                kwargs=Kwargs({"key_1": 1, "key_2": 2}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.GET_STATE,
                # skip content: ledger_id
                contract_id="some str",
                contract_address="some str",
                callable="some str",
                kwargs=Kwargs({"key_1": 1, "key_2": 2}),
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.STATE,
                # skip content: state
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                # skip content: raw_transaction
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.RAW_MESSAGE,
                # skip content: raw_message
            ),
            ContractApiMessage(
                performative=ContractApiMessage.Performative.ERROR,
                # skip content: code
                message=["some str"],
                data=b"some_bytes",
            ),
        ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module compares P2PLibp2p go code in open-aea and open-acn"""

# pylint: skip-file

import difflib
import filecmp
import logging
import os
import tempfile
from collections import Counter, namedtuple
from pathlib import Path
from typing import Any, Iterable, List

import git
import pytest

from aea.test_tools.utils import remove_test_directory

from packages.valory.connections.p2p_libp2p.tests.base import libp2p_log_on_failure_all


PACKAGE = "packages.valory.connections.p2p_libp2p.libp2p_node"
LIBP2P_NODE_DIRECTORY = str(Path(__file__).parent.parent.absolute() / "libp2p_node")

# if testing locally: tmp_dir = "../open-acn/"  # WARNING: will be `/` prefix missing still
ACN_GITHUB_URL = "https://github.com/valory-xyz/open-acn/"

FilePaths = namedtuple("FilePaths", "abs_aea abs_acn rel_aea rel_acn")


def get_all_file_paths(directory: str, extension: str = "") -> List[str]:
    """
    Get all nested files from a directory with a specific extension.

    usage: get_all_file_paths("packages/valory/skills")
    :param directory: Directory to iterate
    :param extension: File extention
    :return: A list of files matching the extention
    """

    return [
        os.path.join(root, file)
        for root, _, files in os.walk(directory)
        for file in files
        if file.endswith(extension)
    ]


def get_relative_file_paths(root: str, *abs_paths: Any) -> List[str]:
    """Remove the root directory from the absolute paths"""

    return [abs_path.split(root).pop() for abs_path in abs_paths]


@pytest.fixture(scope="class")
def acn_repo_dir() -> Any:
    """We keep the ACN repo around for all tests."""

    tmp_dir = tempfile.mkdtemp()
    acn_repo = git.Repo.clone_from(ACN_GITHUB_URL, tmp_dir)
    yield tmp_dir, acn_repo
    remove_test_directory(tmp_dir)


@pytest.fixture(scope="class")
def go_file_paths(acn_repo_dir: Any) -> FilePaths:
    """Get go file paths"""

    tmp_dir, acn_repo = acn_repo_dir
    abs_aea = get_all_file_paths(LIBP2P_NODE_DIRECTORY, ".go")
    abs_acn = get_all_file_paths(tmp_dir, ".go")
    rel_aea = get_relative_file_paths(LIBP2P_NODE_DIRECTORY, *abs_aea)
    rel_acn = get_relative_file_paths(tmp_dir, *abs_acn)
    return FilePaths(abs_aea, abs_acn, rel_aea, rel_acn)


@libp2p_log_on_failure_all
class TestP2PLibp2pGoCodeMatchingOpenACN:
    """
    Test that the open-aea p2p_libp2p go code matches that in open-acn.

    Steps:
    1. download the open-acn repo
    2. ensure all required files are present
    3. use filecmp and difflib to check for differences
    """

    def setup(self) -> Any:  # I don't take extra args :)
        """Set the test up"""

    def test_repo_not_bare(self, acn_repo_dir: Any) -> Any:
        """Check that the repo isn't bare"""

        _, acn_repo = acn_repo_dir
        assert not acn_repo.bare

    def test_unique_relative_path_assumption(self, go_file_paths: FilePaths) -> Any:
        """Check assumption of unique relative paths."""

        def has_duplicates(items: Iterable) -> bool:
            return any(k for k, v in Counter(items).items() if v > 1)

        relative_paths = (go_file_paths.rel_aea, go_file_paths.rel_acn)
        assert not any(map(has_duplicates, relative_paths))

    def test_file_presence(self, go_file_paths: FilePaths) -> Any:
        """Compare both ways to detect missing files."""

        relative_paths = (go_file_paths.rel_aea, go_file_paths.rel_acn)
        aea_filepaths, acn_filepaths = map(set, relative_paths)
        missing_in_acn = aea_filepaths - acn_filepaths
        missing_in_aea = acn_filepaths - aea_filepaths

        assert not missing_in_acn and not missing_in_aea

    def test_content_equal(self, go_file_paths: FilePaths) -> Any:
        """Compare file content, report differences."""

        differences = {}
        zipper = zip(sorted(go_file_paths.abs_aea), sorted(go_file_paths.abs_acn))
        for aea_file, acn_file in zipper:
            # this check is fast, but provides no insight when false
            if filecmp.cmp(aea_file, acn_file):
                continue

            # to facilitate debugging sessions
            aea_content = open(aea_file).readlines()
            acn_content = open(acn_file).readlines()
            detected = difflib.unified_diff(aea_content, acn_content)
            file_name = aea_file.split(LIBP2P_NODE_DIRECTORY).pop()
            differences[file_name] = "".join(detected)

        if differences:
            report: str = "\n".join(f">> {k}:{v}" for k, v in differences.items())
            logging.error(f"Non-matching code:\n{report}")
        assert not bool(differences)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests of the messages module."""

import os
from pathlib import Path
from types import ModuleType

from libs.go.aealite.protocols.acn import v1_0_0 as aealite_acn  # type: ignore

from packages.valory.protocols import acn as package_acn


def test_aealite_protocol_matching():
    """Ensure ACN protocol files are identical on aealite"""

    def get_path(module: ModuleType) -> Path:
        return Path(os.path.sep.join(module.__package__.split(".")))  # type: ignore

    acn_path, aealite_path = map(get_path, (package_acn, aealite_acn))

    readme = (acn_path / "README.md").read_text()
    yaml = (aealite_path / "acn.yaml").read_text()
    assert yaml in readme

    go_package = """option go_package = "libp2p_node/protocols/acn/v1_0_0";\n\n"""
    acn, aealite = ((p / "acn.proto").read_text() for p in (acn_path, aealite_path))
    assert acn == aealite.replace(go_package, "")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests of the ledger connection module."""

import asyncio
import logging
import time
from asyncio import Task
from threading import Thread
from typing import Any, Callable, Dict, FrozenSet, Tuple, Type, cast
from unittest import mock

import pytest
from aea_ledger_ethereum import EthereumCrypto

from aea.common import Address
from aea.configurations.base import ConnectionConfig
from aea.connections.base import ConnectionStates
from aea.crypto.base import LedgerApi
from aea.exceptions import AEAEnforceError
from aea.helpers.async_utils import AsyncState
from aea.mail.base import Envelope
from aea.multiplexer import Multiplexer
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue, DialogueLabel, Dialogues

from packages.valory.connections.ledger.base import RequestDispatcher
from packages.valory.connections.ledger.connection import LedgerConnection, PUBLIC_ID
from packages.valory.connections.ledger.ledger_dispatcher import (
    LedgerApiRequestDispatcher,
)
from packages.valory.connections.ledger.tests.conftest import make_ledger_api_connection

# pylint: skip-file
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.ledger_api import LedgerApiMessage
from packages.valory.protocols.ledger_api.custom_types import Kwargs


SOME_SKILL_ID = "some/skill:0.1.0"
NON_BLOCKING_TIME = 1
BLOCKING_TIME = 100
TOLERANCE = 1
WAIT_TIME_AMONG_TASKS = 0.1


def dummy_task_wrapper(waiting_time: float, result_message: LedgerApiMessage) -> Task:
    """Create a dummy task that simply waits for a given `waiting_time` and returns a given `result_message`."""

    async def dummy_task(*_: Any) -> LedgerApiMessage:
        """Wait for the given `waiting_time` and return the given `result_message`."""
        await asyncio.sleep(waiting_time)
        return result_message

    task = asyncio.create_task(dummy_task())
    return task


class TestLedgerConnection:
    """Test `LedgerConnection` class."""

    @pytest.mark.asyncio
    async def test_not_hanging(self) -> None:
        """Test that the connection does not hang and that the tasks cannot get blocked."""
        # create configurations for the test, re bocking and non-blocking tasks' waiting times
        assert (
            NON_BLOCKING_TIME + TOLERANCE + WAIT_TIME_AMONG_TASKS < BLOCKING_TIME
        ), "`NON_BLOCKING_TIME + TOLERANCE + WAIT_TIME_AMONG_TASKS` should be less than the `BLOCKING_TIME`."

        # setup a dummy ledger connection
        ledger_connection = LedgerConnection(
            configuration=ConnectionConfig("ledger", "valory", "0.19.0"),
            data_dir="test_data_dir",
        )

        # connect() is called by the multiplexer
        await ledger_connection.connect()

        # once a connection is ready, `receive()` is called by the multiplexer
        receive_task = asyncio.ensure_future(ledger_connection.receive())

        # create a blocking task lasting `BLOCKING_TIME` secs
        blocking_task = dummy_task_wrapper(
            BLOCKING_TIME,
            LedgerApiMessage(
                LedgerApiMessage.Performative.ERROR,
                _body={"data": b"blocking_task"},  # type: ignore
            ),
        )
        blocking_dummy_envelope = Envelope(
            to="test_blocking_to",
            sender="test_blocking_sender",
            message=LedgerApiMessage(
                LedgerApiMessage.Performative.ERROR
            ),  # type: ignore
        )
        with mock.patch.object(
            LedgerConnection, "_schedule_request", return_value=blocking_task
        ):
            await ledger_connection.send(blocking_dummy_envelope)

        # create a non-blocking task lasting `NON_BLOCKING_TIME` secs, after `WAIT_TIME_AMONG_TASKS`
        await asyncio.sleep(WAIT_TIME_AMONG_TASKS)
        normal_task = dummy_task_wrapper(
            NON_BLOCKING_TIME,
            LedgerApiMessage(
                LedgerApiMessage.Performative.ERROR, _body={"data": b"normal_task"}
            ),  # type: ignore
        )
        normal_dummy_envelope = Envelope(
            to="test_normal_to",
            sender="test_normal_sender",
            message=LedgerApiMessage(
                LedgerApiMessage.Performative.ERROR
            ),  # type: ignore
        )
        with mock.patch.object(
            LedgerConnection, "_schedule_request", return_value=normal_task
        ):
            await ledger_connection.send(normal_dummy_envelope)

        # sleep for `NON_BLOCKING_TIME + TOLERANCE`
        await asyncio.sleep(NON_BLOCKING_TIME + TOLERANCE)

        # the normal task should be finished
        assert normal_task.done(), "Normal task should be done at this point."

        # the `receive_task` should be done, and not await for the `blocking_task`
        assert receive_task.done(), "Receive task is blocked by blocking task!"

        # the blocking task should not be done
        assert not blocking_task.done(), "Blocking task should be still running."
        # cancel remaining task before ending test
        blocking_task.cancel()


class TestRequestDispatcher:
    """Test `RequestDispatcher` class."""

    dispatcher: RequestDispatcher
    loop: asyncio.AbstractEventLoop

    def setup(self) -> None:
        """Setup test vars."""
        logger = logging.getLogger(type(self).__class__.__name__)
        state = AsyncState(ConnectionStates.connected)
        self.loop = asyncio.get_event_loop()
        self.dispatcher = DummyRequestDispatcher(
            logger=logger, connection_state=state, loop=self.loop
        )

    def dummy_func(self, sleep: float) -> bool:
        """A dummy function that sleeps and returns True."""
        time.sleep(sleep)
        return True

    async def dummy_async_func(self, sleep: float) -> bool:
        """A dummy async function that sleeps and returns True."""
        await asyncio.sleep(sleep)
        return True

    @pytest.mark.asyncio
    async def test_wait_for_happy_path(self) -> None:
        """Tests that wait_for works when timeout is bigger than execution time of callable."""
        should_finish_in = 0.5
        tolerance = 0.5
        timeout = should_finish_in + tolerance

        return_value = await self.dispatcher.wait_for(
            lambda: self.dummy_func(should_finish_in), timeout=timeout
        )
        assert return_value, "dummy_func() should've returned True"

    @pytest.mark.asyncio
    async def test_wait_for_raise_excp(self) -> None:
        """Tests that wait_for works when timeout is less than execution time of callable."""
        timeout = 0.25
        timeout_increase = 0.5
        should_finish_in = timeout + timeout_increase

        with pytest.raises(asyncio.TimeoutError):
            await self.dispatcher.wait_for(
                lambda: self.dummy_func(should_finish_in), timeout=timeout
            )

    @pytest.mark.asyncio
    async def test_wait_for_coroutine(self) -> None:
        """Tests that an error is thrown when a coroutine is passed."""
        should_finish_in = 0.0  # this value is irrelevant to the result of the test
        with pytest.raises(AEAEnforceError):
            await self.dispatcher.wait_for(self.dummy_async_func, should_finish_in)


class DummyLedgerApiMessage(LedgerApiMessage):
    """Implement a dummy `LedgerApiMessage`, which contains performatives for the normal and blocking tasks."""

    class Performative(Message.Performative):
        """Performatives for the ledger_api protocol."""

        NORMAL = "normal"
        BLOCKING = "blocking"
        GET_NORMAL = "get_normal"
        GET_BLOCKING = "get_blocking"

    def __init__(
        self,
        performative: Performative,
        dialogue_reference: Tuple[str, str] = ("", ""),
        message_id: int = 1,
        target: int = 0,
        **kwargs: Any,
    ):
        """Initialise an instance of `DummyLedgerApiMessage`."""
        Message.__init__(
            self,
            dialogue_reference=dialogue_reference,
            message_id=message_id,
            target=target,
            performative=DummyLedgerApiMessage.Performative(performative),
            **kwargs,
        )

    def _is_consistent(self) -> bool:
        """Dummy consistency checks."""
        return True


class DummyLedgerApiDialogue(Dialogue):
    """Implement a dummy `LedgerApiDialogue`."""

    INITIAL_PERFORMATIVES: FrozenSet[Message.Performative] = frozenset(
        {
            DummyLedgerApiMessage.Performative.GET_NORMAL,
            DummyLedgerApiMessage.Performative.GET_BLOCKING,
        }
    )
    TERMINAL_PERFORMATIVES: FrozenSet[Message.Performative] = frozenset(
        {
            DummyLedgerApiMessage.Performative.NORMAL,
            DummyLedgerApiMessage.Performative.BLOCKING,
        }
    )
    VALID_REPLIES: Dict[Message.Performative, FrozenSet[Message.Performative]] = {
        DummyLedgerApiMessage.Performative.GET_NORMAL: frozenset(
            {DummyLedgerApiMessage.Performative.NORMAL}
        ),
        DummyLedgerApiMessage.Performative.GET_BLOCKING: frozenset(
            {DummyLedgerApiMessage.Performative.BLOCKING}
        ),
    }

    class Role(Dialogue.Role):
        """This class defines the agent's role in a ledger_api dialogue."""

        AGENT = "agent"

    class EndState(Dialogue.EndState):
        """This class defines the end states of a ledger_api dialogue."""

        SUCCESSFUL = 0

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: Dialogue.Role,
        message_class: Type[DummyLedgerApiMessage] = DummyLedgerApiMessage,
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for
        :param message_class: the message class used
        """
        Dialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            message_class=message_class,
            self_address=self_address,
            role=role,
        )


class DummyLedgerApiDialogues(Dialogues):
    """Implement a dummy `LedgerApiDialogues`."""

    END_STATES = frozenset({DummyLedgerApiDialogue.EndState.SUCCESSFUL})

    def __init__(self, self_address: Address, **kwargs: Any) -> None:
        """Initialize dialogues."""

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> Dialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message"""
            return DummyLedgerApiDialogue.Role.AGENT

        Dialogues.__init__(
            self,
            self_address=self_address,
            end_states=cast(FrozenSet[Dialogue.EndState], self.END_STATES),
            role_from_first_message=role_from_first_message,
            message_class=DummyLedgerApiMessage,
            dialogue_class=DummyLedgerApiDialogue,
        )


class DummyRequestDispatcher(RequestDispatcher):
    """Implement a dummy request dispatcher."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the dispatcher."""
        super().__init__(*args, **kwargs)
        self.block = True
        self._ledger_api_dialogues = DummyLedgerApiDialogues(
            LedgerConnection.connection_id.__str__()
        )

    @staticmethod
    def get_normal(
        _api: LedgerApi,
        message: DummyLedgerApiMessage,
        dialogue: DummyLedgerApiDialogue,
    ) -> DummyLedgerApiMessage:
        """
        Send the request 'get_normal'.

        :param _api: the API object.
        :param message: the Ledger API message
        :param dialogue: the Ledger API dialogue
        :return: response Ledger API message
        """
        time.sleep(NON_BLOCKING_TIME)

        return cast(
            DummyLedgerApiMessage,
            dialogue.reply(
                performative=DummyLedgerApiMessage.Performative.NORMAL,  # type: ignore
                target_message=message,
                data=b"normal_task",
                ledger_id=message.ledger_id,
            ),
        )

    def get_blocking(
        self,
        _api: LedgerApi,
        message: DummyLedgerApiMessage,
        dialogue: DummyLedgerApiDialogue,
    ) -> DummyLedgerApiMessage:
        """
        Send the request 'get_blocking'.

        :param _api: the API object.
        :param message: the Ledger API message
        :param dialogue: the Ledger API dialogue
        :return: response Ledger API message
        """
        while self.block:
            time.sleep(0.05)

        return cast(
            DummyLedgerApiMessage,
            dialogue.reply(
                performative=DummyLedgerApiMessage.Performative.BLOCKING,  # type: ignore
                target_message=message,
                data=b"blocking_task",
                ledger_id=message.ledger_id,
            ),
        )

    @property
    def dialogues(self) -> Dialogues:
        """Dummy dialogues property."""
        return self._ledger_api_dialogues

    def get_error_message(
        self, exception: Exception, api: LedgerApi, message: Message, dialogue: Dialogue
    ) -> Message:
        """Dummy `get_error_message`."""
        return cast(
            DummyLedgerApiMessage,
            dialogue.reply(
                performative=DummyLedgerApiMessage.Performative.ERROR,  # type: ignore
                target_message=message,
                data=f"{exception}".encode("utf-8"),
                ledger_id=cast(DummyLedgerApiMessage, message).ledger_id,
            ),
        )

    def get_ledger_id(self, message: Message) -> str:
        """Dummy `get_ledger_id`."""
        if not isinstance(message, DummyLedgerApiMessage):  # pragma: nocover
            raise ValueError("argument is not a `DummyLedgerApiMessage` instance.")
        return message.ledger_id

    def get_chain_id(self, message: Message) -> str:
        """Extract the chain id from the message."""
        return self.get_ledger_id(message=message)


class LedgerConnectionWithDummyDispatcher(LedgerConnection):
    """An extended `LedgerConnection` which utilizes the `DummyDispatcher`."""

    async def connect(self) -> None:
        """Set up the connection."""
        if self.is_connected:  # pragma: nocover
            return

        self.state = ConnectionStates.connecting

        self._ledger_dispatcher = DummyRequestDispatcher(  # type: ignore
            logger=self.logger,
            connection_state=self._state,
            loop=self.loop,
            api_configs=self.api_configs,
            retry_attempts=self.request_retry_attempts,
            retry_timeout=self.request_retry_timeout,
        )

        self._response_envelopes = asyncio.Queue()
        self.state = ConnectionStates.connected


class TestLedgerConnectionWithMultiplexer:
    """Test `LedgerConnection` class, using the multiplexer."""

    running_loop: asyncio.AbstractEventLoop
    thread_loop: Thread
    multiplexer: Multiplexer
    ledger_connection: LedgerConnectionWithDummyDispatcher
    make_ledger_connection_callable: Callable = make_ledger_api_connection
    ledger_api_dialogues: DummyLedgerApiDialogues

    @classmethod
    def setup_class(cls) -> None:
        """Setup the test class."""
        # set up a multiplexer with the required ledger connection
        cls.running_loop = asyncio.new_event_loop()
        cls.thread_loop = Thread(target=cls.running_loop.run_forever)
        cls.thread_loop.start()
        cls.multiplexer = Multiplexer(
            [
                LedgerConnectionWithDummyDispatcher(
                    configuration=ConnectionConfig("ledger", "valory", "0.19.0"),
                    data_dir="test_data_dir",
                )
            ],
            loop=cls.running_loop,
        )
        cls.multiplexer.connect()
        # the ledger connection's connect() is called by the multiplexer
        # once a connection is ready, `receive()` is called by the multiplexer
        cls.ledger_connection = cast(
            LedgerConnectionWithDummyDispatcher, cls.multiplexer.connections[0]
        )
        assert cls.ledger_connection._ledger_dispatcher is not None
        cls.ledger_api_dialogues = cast(
            DummyLedgerApiDialogues,
            cls.ledger_connection._ledger_dispatcher._ledger_api_dialogues,
        )

    @classmethod
    def teardown_class(cls) -> None:
        """Tear down the multiplexer."""
        cls.ledger_connection._ledger_dispatcher.block = False  # type: ignore
        if cls.multiplexer.is_connected:
            cls.multiplexer.disconnect()
        cls.running_loop.call_soon_threadsafe(cls.running_loop.stop)
        cls.thread_loop.join()

    def create_ledger_dialogues(
        self, blocking: bool = True
    ) -> Tuple[DummyLedgerApiMessage, DummyLedgerApiDialogues]:
        """Create a dialogue."""
        if blocking:
            performative = DummyLedgerApiMessage.Performative.GET_BLOCKING
            _callable = "get_blocking"
        else:
            performative = DummyLedgerApiMessage.Performative.GET_NORMAL
            _callable = "get_normal"

        return cast(
            Tuple[DummyLedgerApiMessage, DummyLedgerApiDialogues],
            self.ledger_api_dialogues.create(
                counterparty=str(self.ledger_connection.connection_id),
                performative=performative,  # type: ignore
                ledger_id=EthereumCrypto.identifier,
                callable=_callable,
                args=(),
                kwargs=Kwargs({}),
            ),
        )

    @staticmethod
    def create_envelope(request: DummyLedgerApiMessage) -> Envelope:
        """Create a dummy envelope."""
        return Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

    @pytest.mark.asyncio
    async def test_not_hanging_with_multiplexer(self) -> None:
        """Test that the connection does not hang and that the tasks cannot get blocked, using the multiplexer."""
        # create configurations for the test, re bocking and non-blocking tasks' waiting times
        assert (
            NON_BLOCKING_TIME + TOLERANCE + WAIT_TIME_AMONG_TASKS < BLOCKING_TIME
        ), "`NON_BLOCKING_TIME + TOLERANCE + WAIT_TIME_AMONG_TASKS` should be less than the `blocking_time`."
        assert self.ledger_connection._ledger_dispatcher.block  # type: ignore

        # create a blocking task lasting `BLOCKING_TIME` secs
        request, _ = self.create_ledger_dialogues()
        request._sender = SOME_SKILL_ID
        blocking_dummy_envelope = TestLedgerConnectionWithMultiplexer.create_envelope(
            request
        )
        self.multiplexer.put(blocking_dummy_envelope)

        # create a non-blocking task lasting `NON_BLOCKING_TIME` secs, after `WAIT_TIME_AMONG_TASKS`
        await asyncio.sleep(WAIT_TIME_AMONG_TASKS)

        request, _ = self.create_ledger_dialogues(blocking=False)
        request._sender = SOME_SKILL_ID
        normal_dummy_envelope = TestLedgerConnectionWithMultiplexer.create_envelope(
            request
        )
        self.multiplexer.put(normal_dummy_envelope)

        # the response envelopes of the ledger connection should be empty
        assert (
            self.ledger_connection.response_envelopes.empty()
        ), "The response envelopes of the ledger connection should be empty."
        # `receive()` should not be done,
        # and multiplexer's `_receiving_loop` should be still running and have an empty `in_queue`
        assert (
            len(self.multiplexer.in_queue.queue) == 0
        ), "The multiplexer's `in_queue` should not contain anything."

        # sleep for `NON_BLOCKING_TIME + TOLERANCE`
        await asyncio.sleep(NON_BLOCKING_TIME + TOLERANCE)

        # `receive()` should be done,
        # and multiplexer's `_receiving_loop` should have put the `normal_dummy_envelope` in the `in_queue`
        envelope = self.multiplexer.get(block=True)
        assert envelope is not None
        message = envelope.message
        assert isinstance(message, DummyLedgerApiMessage)
        assert (
            message.data == b"normal_task"
        ), "Normal task should be the first item in the multiplexer's `in_queue`."
        assert self.ledger_connection._ledger_dispatcher.block  # type: ignore


@pytest.mark.asyncio
async def test_request_dispatcher_run_async() -> None:
    """Test request dispatcher run async."""
    dispatcher = LedgerApiRequestDispatcher(
        logger=mock.Mock(), connection_id=PUBLIC_ID, connection_state=AsyncState()
    )
    assert (
        await dispatcher.run_async(
            lambda x, y, z: 12, mock.Mock(), mock.Mock(), mock.Mock()  # type: ignore
        )
        == 12
    )

    def raise_exc(*args):  # type: ignore
        raise ValueError("expected")

    dialogue = mock.Mock()
    dialogue.reply = mock.Mock(return_value=124)
    assert (
        await dispatcher.run_async(
            raise_exc, mock.Mock(), mock.Mock(), dialogue=dialogue
        )
        == 124
    )
    dialogue.reply.assert_called_once_with(
        performative=mock.ANY,
        target_message=mock.ANY,
        code=500,
        message="expected",
        data=b"",
    )


def test_ledger_connection_exceptions() -> None:
    """Test exception."""
    configuration = mock.Mock()
    configuration.public_id = LedgerConnection.connection_id
    connection = LedgerConnection(data_dir="", configuration=configuration)
    with pytest.raises(
        ValueError,
        match="`asyncio.Queue` for `_response_envelopes` not set. Is the ledger connection active?",
    ):
        connection.response_envelopes

    envelope = mock.Mock()
    envelope.protocol_specification_id = ContractApiMessage.protocol_specification_id
    connection._contract_dispatcher = mock.Mock()
    connection._contract_dispatcher.dispatch.return_value = 12
    assert connection._schedule_request(envelope) == 12



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module sets up test environment for erc1155_client skill."""
# pylint: skip-file

from pathlib import Path
from typing import cast

from aea.helpers.search.models import (
    Attribute,
    Constraint,
    ConstraintType,
    DataModel,
    Description,
    Query,
)
from aea.helpers.transaction.base import RawMessage, RawTransaction, Terms
from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.erc1155_client.behaviours import SearchBehaviour
from packages.fetchai.skills.erc1155_client.dialogues import (
    ContractApiDialogues,
    DefaultDialogues,
    FipaDialogues,
    LedgerApiDialogues,
    OefSearchDialogues,
    SigningDialogues,
)
from packages.fetchai.skills.erc1155_client.handlers import (
    ContractApiHandler,
    FipaHandler,
    LedgerApiHandler,
    OefSearchHandler,
    SigningHandler,
)
from packages.fetchai.skills.erc1155_client.strategy import Strategy
from packages.open_aea.protocols.signing.message import SigningMessage
from packages.valory.protocols.contract_api.custom_types import Kwargs
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage


PACKAGE_DIR = Path(__file__).parent.parent


class ERC1155ClientTestCase(BaseSkillTestCase):
    """Sets the erc1155_client class up for testing."""

    path_to_skill = PACKAGE_DIR

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        cls.location = {"longitude": 0.1270, "latitude": 51.5194}
        cls.search_query = {
            "search_key": "seller_service",
            "search_value": "erc1155_contract",
            "constraint_type": "==",
        }
        cls.search_radius = 5.0
        config_overrides = {
            "models": {
                "strategy": {
                    "args": {
                        "location": cls.location,
                        "search_query": cls.search_query,
                        "search_radius": cls.search_radius,
                    }
                }
            },
        }

        super().setup_class(config_overrides=config_overrides)

        # behaviours
        cls.search_behaviour = cast(
            SearchBehaviour, cls._skill.skill_context.behaviours.search
        )

        # dialogues
        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.default_dialogues = cast(
            DefaultDialogues, cls._skill.skill_context.default_dialogues
        )
        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )
        cls.oef_search_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )

        # handlers
        cls.fipa_handler = cast(FipaHandler, cls._skill.skill_context.handlers.fipa)
        cls.oef_search_handler = cast(
            OefSearchHandler, cls._skill.skill_context.handlers.oef_search
        )
        cls.contract_api_handler = cast(
            ContractApiHandler, cls._skill.skill_context.handlers.contract_api
        )
        cls.signing_handler = cast(
            SigningHandler, cls._skill.skill_context.handlers.signing
        )
        cls.ledger_api_handler = cast(
            LedgerApiHandler, cls._skill.skill_context.handlers.ledger_api
        )

        # models
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)

        cls.logger = cls._skill.skill_context.logger

        # mocked objects
        cls.ledger_id = "some_ledger_id"
        cls.contract_id = "some_contract_id"
        cls.contract_address = "some_contract_address"
        cls.callable = "some_callable"
        cls.body = {"some_key": "some_value"}
        cls.kwargs = Kwargs(cls.body)
        cls.address = "some_address"
        cls.mocked_terms = Terms(
            cls.ledger_id,
            cls._skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        cls.mocked_query = Query(
            [Constraint("some_attribute_name", ConstraintType("==", "some_value"))],
            DataModel(
                "some_data_model_name",
                [
                    Attribute(
                        "some_attribute_name",
                        str,
                        False,
                        "Some attribute descriptions.",
                    )
                ],
            ),
        )
        cls.mocked_proposal = Description(
            {
                "contract_address": "some_contract_address",
                "token_id": "123456",
                "trade_nonce": "876438756348568",
                "from_supply": "543",
                "to_supply": "432",
                "value": "67",
            }
        )
        cls.mocked_raw_tx = (RawTransaction(cls.ledger_id, {"some_key": "some_value"}),)
        cls.mocked_raw_msg = RawMessage(cls.ledger_id, b"some_body")

        # list of messages
        cls.list_of_fipa_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": cls.mocked_query}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": cls.mocked_proposal}
            ),
        )
        cls.list_of_oef_search_messages = (
            DialogueMessage(
                OefSearchMessage.Performative.SEARCH_SERVICES,
                {"query": cls.mocked_query},
            ),
        )
        cls.list_of_contract_api_messages = (
            DialogueMessage(
                ContractApiMessage.Performative.GET_RAW_MESSAGE,
                {
                    "ledger_id": cls.ledger_id,
                    "contract_id": cls.contract_id,
                    "contract_address": cls.contract_address,
                    "callable": cls.callable,
                    "kwargs": cls.kwargs,
                },
            ),
        )
        cls.list_of_signing_messages = (
            DialogueMessage(
                SigningMessage.Performative.SIGN_MESSAGE,
                {"terms": cls.mocked_terms, "raw_message": cls.mocked_raw_msg},
            ),
        )
        cls.list_of_ledger_api_messages = (
            DialogueMessage(
                LedgerApiMessage.Performative.GET_BALANCE,
                {"ledger_id": cls.ledger_id, "address": "some_address"},
            ),
        )
